unit ColorPicker;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Buttons;

type
  TColorPopup = class;

  TColorPickButton = class(TSpeedButton)
  private
    FSelected: TColor;        //ANeBuȐFA̐Fŕ\B
    FOnSelect: TNotifyEvent;  //IςɌĂ΂Cxg
    FOnDropDown: TNotifyEvent;
    FColorPopup: TColorPopup;
    FDropDown: Boolean;
    procedure SetSelected(const Value:TColor);
    procedure SetDropDown(const Value: Boolean);
  protected
    procedure ShowPopup;
    procedure Paint;  override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click;  override;
    property DropDown: Boolean read FDropDown write SetDropDown;
  published
    property Selected: TColor read FSelected write SetSelected;
    property OnSelect:TNotifyEvent read FOnSelect write FOnSelect;
    property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;
  end;

  TColorPopup = class(TCustomControl)
  private
    FCellSize: Integer;       //Z̑傫@cƉƂ
    FMargin: Integer;         //Z̗]    ㉺EƂ
    FSpacing: Integer;        //EChẼXy[X
    FColCount: Integer;       //s̐
    FRowCount: Integer;       //̐
    FSelectedIndex: Integer;  //IꂽZ̃CfbNX
    FHoverIndex: Integer;     //}EXz[o[ĂZ̃CfbNX
    FSelectedColor: TColor;   //IĂF
    FParentButton: TColorPickButton;
    function GetCellRect(Index: Integer; var Rect: TRect): Boolean;
    function GetRow(Index: Integer): Integer;
    function GetCol(Index: Integer): Integer;
    function GetIndex(Row, Col: Integer): Integer;
    procedure DrawCell(Index: Integer);
    procedure EndSelection(Cancel: Boolean);
    procedure SetSelectedColor(const Value: TColor);
    procedure SelectColor(Color: TColor);
    procedure ChangeSelection(NewSelection: Integer);
    function SelectionFromPoint(P: TPoint): Integer;
    procedure ChangeHoverSelection(Index: Integer);
    procedure InvalidateCell(Index: Integer);
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
    procedure CNKeyDown(var Message: TWMKeyDown); message CN_KEYDOWN;
    procedure WMActivateApp(var Message: TWMActivateApp); message WM_ACTIVATEAPP;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property ParentButton: TColorPickButton read FParentButton;
  published
    property SelectedColor: TColor read FSelectedColor write SetSelectedColor;
  end;

procedure Register;

implementation

uses
  System.Types;

{$R ColorPicker.res}
{$R ColorPicker.dcr}

const
   DefaultColorCount = 40;

   DefaultColors : array[0..DefaultColorCount - 1] of TColor
     =($000000, $003399, $003333, $003300, $663300, $800000, $993333, $333333,
       $000080, $0066FF, $008080, $008000, $808000, $FF0000, $996666, $808080,
       $0000FF, $0099FF, $00CC99, $669933, $CCCC33, $FF6633, $800080, $969696,
       $FF00FF, $00CCFF, $00FFFF, $00FF00, $FFFF00, $FFCC00, $663399, $C0C0C0,
       $CC99FF, $99CCFF, $99FFFF, $CCFFCC, $FFFFCC, $FFCC99, $FF99CC, $FFFFFF);

   ColorNames : array[0..DefaultColorCount - 1] of string
     =('','','I[u','Z','Zΐ','Z','','80% DF',
       'Z','IW','Z','','','','u[O[','50% DF',
       '','WIW','C','V[O[','ANA','W','','40% DF',
       'sN','','','邢','F','XJCu[','v','25% DF',
       '[Y','x[W','','','F','y[u[','x_[','');

procedure Register;
begin
  RegisterComponents('Sada''s', [TColorPickButton]);
end;

{ TColorPickCombo }

constructor TColorPickButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height :=22;
  Width := 45;
  FSelected := clBlack;
  ShowHint := True;
  Height :=21;
  Width := 45;
  ShowHint := True;
  Layout := blGlyphRight;
  Margin := 0;
  //Glyph := TBitmap.Create; ͕KvȂB
  //ColorPicker.res Ɋ܂܂摜 ARROWDOWN ǂݎB
  Glyph.LoadFromResourceName(HInstance, 'ARROWDOWN');
 end;

destructor TColorPickButton.Destroy;
begin
  inherited Destroy;
end;

procedure TColorPickButton.ShowPopup;
var
  P:TPoint;
begin
  P.X:=0;
  P.Y:=Height;
  P:=ClientToScreen(P);
  // |bvAbvEChE̐
  FColorPopup := TColorPopup.Create(Self);
  with FColorPopup do
  begin
    FParentButton := Self;
    if P.X + Width > Screen.Width then
      P.X := Screen.Width - Width;
    if P.Y + Height > Screen.Height then
      P.Y := P.Y - Height - Self.Height;
    SetWindowPos(Handle, HWND_TOPMOST, P.X, P.Y, Width, Height, SWP_SHOWWINDOW);
    //EndSelection ܂ł̊ԁAׂẴ}EXCxgߊl
    SetCapture(Handle);
  end;
end;

procedure TColorPickButton.Click;
begin
  if Enabled then
  begin
    if not FDropDown then
    begin
       Update;
       DropDown := True;
    end;
  end;
end;

procedure TColorPickButton.Paint;
var
  R: TRect;
begin
  inherited;
  R := Rect(5,4,20,Height-4);
  Canvas.Pen.Color := clGray;
  Canvas.Brush.Color := FSelected;
  Canvas.Rectangle(R);
end;


procedure TColorPickButton.SetSelected(const Value: TColor);
begin
  if Value <> FSelected then
  begin
    FSelected := Value;
    if Assigned(FOnSelect) then FOnSelect(Self);
    invalidate;
  end;
end;

procedure TColorPickButton.SetDropDown(const Value: Boolean);
begin
  if FDropDown <> Value then
  begin
    FDropDown := Value;
    if FDropDown then
    begin
      ShowPopup;
      FColorPopup.SelectedColor := FSelected;
    end
    else
    begin
      ShowWindow(FColorPopup.Handle, SW_HIDE);
      FreeAndNil(FColorPopup);
    end;
    if Assigned(FOnDropDown) then FOnDropDown(Self);
    Invalidate;
  end;
  FDropDown := Value;
end;

{ TColorPopup }

constructor TColorPopup.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := ControlStyle - [csAcceptsControls] + [csOpaque];
  FColCount := 8;
  FRowCount := 5;
  FCellSize := 18;
  FSpacing := 4;
  FMargin := 2;
  FHoverIndex := -1;
  FSelectedIndex := -1;
  Width := FColCount * FCellSize + 2 * FSpacing;
  Height := FRowCount * FCellSize + 2 * FSpacing;
  ShowHint := True;
end;

destructor TColorPopup.Destroy;
begin
  inherited;
end;

procedure TColorPopup.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    WndParent := GetDesktopWindow;
    Style := WS_CLIPSIBLINGS or WS_CHILD;
    ExStyle := WS_EX_TOPMOST or WS_EX_TOOLWINDOW;
    WindowClass.Style := Params.WindowClass.Style or CS_DROPSHADOW;
   end;
end;

procedure TColorPopup.DrawCell(Index: Integer);
var
  R: TRect;
begin
  if GetCellRect(Index, R) then
  begin
    if (FSelectedIndex = Index) and (FHoverIndex <> Index) then
      Canvas.Brush.Color := clHighlight
    else
      Canvas.Brush.Color := clBtnFace;
    Canvas.Pen.Style := psSolid;
    Canvas.FillRect(R);
    Canvas.Pen.Style := psSolid;
    if (FSelectedIndex = Index) then
    begin
      Canvas.Pen.Color := clGray;
      Canvas.Rectangle(R);
    end
    else if FHoverIndex = Index then  //}EXĂ
    begin
      Canvas.Pen.Color := clRed;
      Canvas.Rectangle(R);
    end;

    Canvas.Pen.Color := clGray;
    Canvas.Brush.Color := DefaultColors[Index];
    // Z̐FœhԂ
    InflateRect(R, -FMargin, -FMargin);
    Canvas.Rectangle(R);
  end;
end;

procedure TColorPopup.ChangeHoverSelection(Index: Integer);
begin
  InvalidateCell(FHoverIndex);

  FHoverIndex := Index;
  InvalidateCell(FHoverIndex);
  Update;
end;

procedure TColorPopup.ChangeSelection(NewSelection: Integer);
begin
  if NewSelection <> -1 then
  begin
    if FSelectedIndex <> -1 then
      InvalidateCell(FSelectedIndex);
    FSelectedIndex := NewSelection;
    if FSelectedIndex <> -1 then
      InvalidateCell(FSelectedIndex);
  end;
end;

procedure TColorPopup.CMHintShow(var Message: TMessage);
begin
  with TCMHintShow(Message) do
  begin
    if not ParentButton.ShowHint then
      Message.Result := 1
    else
    begin
      with HintInfo^ do
      begin
        Result := 0;
        HintStr := '';
        if (FHoverIndex >= 0) and (FHoverIndex < DefaultColorCount) then
        begin
          GetCellRect(FHoverIndex, CursorRect);
          HintStr := ColorNames[FHoverIndex];
        end;
      end;
    end;
  end;
end;

//L[{[h͂̏BGridControl܂˂B
procedure TColorPopup.CNKeyDown(var Message: TWMKeyDown);
var
  Row, Col: Integer;
begin
  inherited;

  if FHoverIndex <> -1 then
  begin
    Row := GetRow(FHoverIndex);
    Col := GetCol(FHoverIndex);
  end
  else
  begin
    Row := GetRow(FSelectedIndex);
    Col := GetCol(FSelectedIndex);
  end;

  case Message.CharCode of
    VK_DOWN:
      begin
        Inc(Row);
        if Row >= FRowCount then
          Dec(Row);
        ChangeHoverSelection(GetIndex(Row, Col));
        Message.Result := 1;
      end;
    VK_UP:
      begin
        Dec(Row);
        if Row < 0 then
          Row := 0;
        ChangeHoverSelection(GetIndex(Row, Col));
        Message.Result := 1;
      end;
    VK_RIGHT:
      begin
        Inc(Col);
        if Col = FColCount then
          Dec(Col);
        ChangeHoverSelection(GetIndex(Row, Col));
        Message.Result := 1;
      end;
    VK_LEFT:
      begin
        Dec(Col);
        if Col < 0 then
          Col := 0;
        ChangeHoverSelection(GetIndex(Row, Col));
        Message.Result := 1;
      end;

    VK_ESCAPE:
      begin
        EndSelection(True);
        Message.Result := 1;
      end;

    VK_RETURN, VK_SPACE:
      begin
        FSelectedIndex := FHoverIndex;
        EndSelection(False);
        Message.Result := 1;
      end;
  end;

end;

//Isׂ̏IBEChEB
procedure TColorPopup.EndSelection(Cancel: Boolean);
begin
  with ParentButton do
  begin
    ReleaseCapture;
    DropDown := False;
    if not Cancel then
    begin
      FSelectedColor :=  DefaultColors[FSelectedIndex];
      Selected := FSelectedColor;
    end;
  end;
end;

function TColorPopup.GetCellRect(Index: Integer; var Rect: TRect): Boolean;
begin
  if (Index < 0) or (Index >= DefaultColorCount) then
    Result := False
  else
  begin
    Rect.Left := FSpacing + GetCol(Index) * FCellSize;
    Rect.Top := FSpacing + GetRow(Index) * FCellSize;
    Rect.Right := Rect.Left + FCellSize;
    Rect.Bottom := Rect.Top + FCellSize;
    Result := True;
  end;
end;

function TColorPopup.GetIndex(Row, Col: Integer): Integer;
begin
  Result := Row * FColCount + Col;
end;

function TColorPopup.GetCol(Index: Integer): Integer;
begin
  Result := Index mod FColCount;
end;

function TColorPopup.GetRow(Index: Integer): Integer;
begin
  Result := Index div FColCount;
end;

procedure TColorPopup.InvalidateCell(Index: Integer);
var
  R: TRect;
begin
  if GetCellRect(Index, R) then InvalidateRect(Handle, @R, False);
end;

procedure TColorPopup.Paint;
var
  I: Integer;
begin
  inherited;
  Canvas.Pen.Color := clGray;
  Canvas.Rectangle(ClientRect);

  for I := 0 to DefaultColorCount - 1 do
    DrawCell(I);
end;

//Xg^ꂽF̃CfbNXĐݒ肷
procedure TColorPopup.SelectColor(Color: TColor);
var
  I: Integer;
  Found: Boolean;
begin
  Found := False;
  for I := 0 to DefaultColorCount - 1  do
  begin
    if DefaultColors[I] = Color then
    begin
      FSelectedIndex := I;
      FSelectedColor := Color;
      Found := True;
      Break;
    end;
  end;
  if Not Found then
  begin
    FSelectedIndex := -1;
    FSelectedColor := clNone;
  end;
end;

// }EXʒuZ̃CfbNX擾
function TColorPopup.SelectionFromPoint(P: TPoint): Integer;
begin
  Result := -1;
  if (P.X > FSpacing) and (P.X < Width - FSpacing)
    and (P.Y > FSpacing) and (P.Y < Height - FSpacing) then
  begin
    Dec(P.X, FSpacing);
    Dec(P.Y, FSpacing);
    Result := GetIndex(P.Y div FCellSize, P.X div FCellSize);
    if Result > DefaultColorCount then
      Result := -1;
  end;
end;

// FSelectedColorݒ肷B
procedure TColorPopup.SetSelectedColor(const Value: TColor);
begin
  if FSelectedColor <> Value then
  begin
    FSelectedColor := Value;
    SelectColor(Value);
  end;
end;

//Active łȂȂAI߂ĕB
procedure TColorPopup.WMActivate(var Message: TWMActivate);
begin
  inherited;
  if not Message.Active = 0 then
    EndSelection(True);
end;

//̃AvP[VNAI߂ĕB
procedure TColorPopup.WMActivateApp(var Message: TWMActivateApp);
begin
  inherited;
  if not Message.Active  then
    EndSelection(True);
end;

//tH[JXB
procedure TColorPopup.WMKillFocus(var Message: TWMKillFocus);
begin
  inherited;
  ParentButton.DropDown := False;
end;

//}EXAbvbZ[Wnh
//̃Rg[ȊÔƂ́AILZďIB
procedure TColorPopup.WMLButtonUp(var Message: TWMLButtonUp);
var
  NewSelection: Integer;
begin
  inherited;
  if PtInRect(ClientRect, Point(Message.XPos, Message.YPos)) then
  begin
    NewSelection := SelectionFromPoint(Point(Message.XPos, Message.YPos));
    if NewSelection <> -1 then
      ChangeSelection(NewSelection);
    EndSelection(False);
  end
  else
    EndSelection(True);
end;

// }EXʒũZ̃CfbNXFHoverIndexɐݒ肷B
procedure TColorPopup.WMMouseMove(var Message: TWMMouseMove);
var
  NewSelection: Integer;
begin
  inherited;
  NewSelection := SelectionFromPoint(Point(Message.XPos, Message.YPos));
  if NewSelection <> FHoverIndex then
    ChangeHoverSelection(NewSelection);
end;


end.
