{-----------------------------------------------------------------------}
{                                                                       }
{  GPXYGraph jbg                                                   }
{                                                                       }
{  10 Mar. 2020  Generics                                             }
{                                                                       }
{                                                                       }
{                                                                       }
{                                                                       }
{                                              @copyright   S. Sekine   }
{-----------------------------------------------------------------------}

unit GPXYGraph;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  System.Win.ComObj,  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  System.Generics.Collections, System.Math, System.Types, Vcl.ClipBrd, System.UITypes,
  pngimage, GifImg,
  Winapi.GDIPAPI, Winapi.GDIPOBJ;

const
  epsilon: double = 1.0E-20;
  MinimumRange = 1.0E-8;
  MaxTicksCount = 30;
  MaxSubTicksCount = 20;

type

  { Integerli[铮Iz}
  TGraphIntegerList = class(TList<Integer>)
  end;

  TGraphPointList = class(TList<TPoint>)
  public
    function AddXY(X,Y: integer): Integer;
  end;

  TGraphValueList = class(TList<Double>)
  private
    FMinValue: double;
    FMaxValue: double;
    FMinValueIndex: Integer;
    FMaxValueIndex: Integer;
  protected
    procedure PutValue(Index: Integer; Value: double);
  public
    procedure Assign(Source: TGraphValueList);
    function Add(Value: double): Integer;
    procedure ReCalcMinMax;
    property MaxValue: double read FMaxValue;
    property MinValue: double read FMinValue;
    property MinValueIndex: Integer read FMinValueIndex;
    property MaxValueIndex: Integer read FMaxValueIndex;
  end;

  TGraphBooleanList = class(TList<Boolean>)
  end;

const
  MaxDefaultColors = 16;
  ColorPalette : Array[0..MaxDefaultColors-1] of TColor
                  = (clRed, clGreen, clBlue, clFuchsia, clTeal, clNavy,
                     clMaroon, clLime, clOlive, clPurple, clSilver, clAqua,
                     clBlack, clYellow, clWhite, clGray);

type
  TGPXYGraph = class;

  EGPXYGraphError = class(Exception);
  EXYGraphAxisError = class(EGPXYGraphError);
  EXYGraphSeriesError = class(EGPXYGraphError);
  TAxisOrientation = (aoHorizontal, aoVertical);
  TAxisScale = (asNormal, asPi, asPercent, asLog);
  TAxisTitleLoc = (tlHead, tlFoot, tlSide, tlFixed);

  TGPGraphAxis = class(TPersistent)
  private
    {ẽRg[}
    FParentGraph: TGPXYGraph;
    {̌}
    FOrientation: TAxisOrientation;
    {ݒ}
    FAutoStep: boolean;
    FAutoOrigin: boolean;
    FAutoMin: boolean;
    FAutoMax: boolean;
    FAutoDecimal: boolean;

    FVisible: boolean;
    FTicks: boolean;          //̖ڐt          default true
    FLabels: boolean;         //̃xt        default true
    {ڐ͈̔}
    FMax : double;            //ől
    FMin : double;            //ŏl
    FRange: double;           //͈͒l            FMax - FMin
    FLogRange: double;        //ΐڐ͈̔͒l  Log10(FMax/FMin)
    {ڐԊu}
    FTickStep: double;        //ڐԂ̊Ԋu
    FTicksCount: integer;     //ڐ̊Ґ
    FCurrTicksCount: integer; //ڐ̐
    FSubTicksCount: integer;  //ڐ̊Ԃɂ镛ڐ̐
    {ڐ̒li[郊Xg}
    FTickDataList: TGraphValueList;
    FSubTickDataList: TGraphValueList;
    {ڐXg̃f[^}
    {Ɍ鎲Ƃ̌_}
    FOrigin: double;          //̌_̒l

    {`}
    FAxisPen: TPen;           //`y
    FLabelFont: TFont;        //x̃tHg
    FTickLen: integer;        //ڐ̒isNZj
    FSubTickLen: integer;     //̒isNZj
    FExtraLen: integer;       //őEŏl̗]ȒisNZj
    {̑}
    FInverted: boolean;       //őlƍŏlt]
    FAxisScale: TAxisScale;
    FLineThickness: Single;
    {̈ʒuw}
    FAxisBounds: TRect;       //`̈
    FMinPos :integer;         //ŏl̈ʒuiLoX̐Βlj
    FMaxPos : integer;        //ől̈ʒuiLoX̐Βlj
    FStartPos :integer;       //̎n_iLoX̐Βlj
    FEndPos : integer;        //̏I_iLoX̐Βlj
    FPlotSize : integer;      //̎̒isNZj

    {x}
    FLabelSpace: integer;     //xƎƂ̋isNZj
    FLabelCount: integer;     //x̐
    FLabelDecimal: integer;   //x̏_ȉ̌
    FLabelsOnAxis: boolean;   //̌_Ƀxt  default false
    FLabelValueList: TGraphValueList;  //xtl̃Xg
    {^Cg}
    FTitle: string;
    FTitleAngle: integer;     //^Cg̉]px
    FTitleFont: TFont;
    FAxisTitleLoc: TAxisTitleLoc;

    {ANZX\bh}
    procedure SetVisible(const Value: boolean);
    procedure SetAutoDecimal(const Value: boolean);
    procedure SetAutoMin(const Value: boolean);
    procedure SetAutoMax(const Value: boolean);
    procedure SetAutoStep(const Value: boolean);
    procedure SetAutoOrigin(const Value: boolean);
    procedure SetAxisBounds(const Value: TRect);
    procedure SetAxisScale(const Value: TAxisScale);
    procedure SetAxisPen(const Value: TPen);
    procedure SetInverted(const Value: boolean);
    procedure SetLabelDecimal(const Value: integer);
    procedure SetLabelFont(const Value: TFont);
    procedure SetLabels(const Value: boolean);
    procedure SetLabelSpace(const Value: integer);
    procedure SetLabelsOnAxis(const Value: boolean);
    procedure SetMax(const Value: double);
    procedure SetMin(const Value: double);
    procedure SetOrientation(const Value: TAxisOrientation);
    procedure SetOrigin(const Value: double);
    procedure SetSubTicksCount(const Value: integer);
    procedure SetSubTickLen(const Value: integer);
    procedure SetTickLen(const Value: integer);
    procedure SetTicks(const Value: boolean);
    procedure SetTickStep(const Value: double);
    procedure SetTitleFont(const Value: TFont);
    procedure SetTitle(const Value: string);
    procedure SetTitleAngle(const Value: integer);
    procedure SetAxisTitleLoc(const Value: TAxisTitleLoc);
    function GetAxisWidth: integer;
    function GetAxisHeight: integer;
    function GetOriginPos: integer;
    function PiSymbol: string;
    procedure SetTicksCount(const Value: integer);
    procedure SetExtraLen(const Value: integer);
    procedure SetLineThickness(const Value: Single);
  protected
    procedure AddTickData(Value: double);
    procedure AddSubTickData(Value: double);
    function CalcTickStep(Range: double):double;
    procedure AxisChanged(Sender: TObject);
    procedure AdjustSize;
    procedure Error(const Msg:string);
    function GetTitleHeight(Loc: TAxisTitleLoc) :integer;
    function GetTitleWidth(Loc: TAxisTitleLoc) :integer;
    procedure SetTickValues;
    procedure SetLogTickValues;
    function CheckTickStep(Value: double): boolean;
    function CalcOriginValue(AMin, AMax, AStep: double): double;
    procedure SetLabelValues;
    procedure DrawAxisLine(Pos: Integer);
    procedure DrawAxisTicks(Pos: Integer; Reverse: Boolean);
    procedure DrawAxisLabels(Pos: Integer);
    procedure DrawTitle(Pos: Integer);
    procedure PutLabel(AFont: TGPFont; ABrush: TGPBrush;
      Pos: Integer; Value: double);
    procedure DrawTicksLines(List: TGraphValueList; Pos, Len: integer);
    procedure SetAxisRange(const AMin, AMax: double);
    function CreateGPPen(APen: TPen): TGPPen;
    function TitleSize: TSize;
    function TitleHeight: integer;
    function TitleWidth: integer;
    procedure UpdateScale;
    function GetSeriesMin :double;
    function GetSeriesMax :double;
  public
    constructor Create(ParentGraph: TGPXYGraph);
    destructor Destroy; override;
    procedure Init;
    procedure Draw(Pos: Integer; Reverse: Boolean);
    procedure ResetScale;
    function GetLabelHeight(Value: double):integer;
    function GetLabelWidth(Value: double):integer;
    function GetMaxLabelsHeight:integer;
    function GetMaxLabelsWidth: integer;
    function GetOriginMargin: integer;
    function GetPixelsPerUnit:double;
    function GetValuePos(Value: double):integer;
    function GetValueSize(Value: double):integer;
    function GetPosValue(Pos: integer):double;
    function GetValueText(Value: double):string;
    function GetValuePerPixel:double;
    procedure SetMinMax(const AMin, AMax: double);

    property AxisBounds:TRect read FAxisBounds write SetAxisBounds;
    property AxisHeight: integer read GetAxisHeight; { read only }
    property AxisWidth: integer read GetAxisWidth;  { read only }
    property LabelValueList: TGraphValueList read FLabelValueList; { read only }
    property MaxPos: integer read FMaxPos;       { read only }
    property MinPos: integer read FMinPos;       { read only }
    property StartPos: integer read FStartPos;   { read only }
    property EndPos: integer read FEndPos;       { read only }
    property Orientation: TAxisOrientation read FOrientation write SetOrientation;
    property OriginPos: integer read GetOriginPos;  { read only }
    property ParentGraph: TGPXYGraph read FParentGraph;
    property PlotSize: integer read FPlotSize;     { read only }
    property SubTickDataList: TGraphValueList read FSubTickDataList; { read only }
    property TickDataList: TGraphValueList read FTickDataList; { read only }
  published
    property AutoDecimal: boolean read FAutoDecimal write SetAutoDecimal default true;
    property AutoMax: boolean read FAutoMax write SetAutoMax default true;
    property AutoMin: boolean read FAutoMin write SetAutoMin default true;
    property AutoOrigin: boolean read FAutoOrigin write SetAutoOrigin default true;
    property AxisPen:TPen read FAxisPen write SetAxisPen;
    property AxisScale: TAxisScale read FAxisScale write SetAxisScale default asNormal;
    property AutoStep: boolean read FAutoStep write SetAutoStep default true;
    property AxisTitleLoc: TAxisTitleLoc read FAxisTitleLoc write SetAxisTitleLoc default tlHead;
    property ExtraLen: integer read FExtraLen write SetExtraLen default 8;
    property Inverted: boolean read FInverted write SetInverted default false;
    property LabelDecimal: integer read FLabelDecimal write SetLabelDecimal default 0;
    property LabelFont:TFont read FLabelFont write SetLabelFont;
    property Labels: boolean read FLabels write SetLabels default true;
    property LabelsOnAxis: boolean read FLabelsOnAxis write SetLabelsOnAxis default False;
    property LabelSpace: integer read FLabelSpace write SetLabelSpace default 4;
    property Min: double read FMin write SetMin;
    property Max: double read FMax write SetMax;
    property Origin: double read FOrigin write SetOrigin;
    property SubTickLen: integer read FSubTickLen write SetSubTickLen default 3;
    property SubTicksCount: integer read FSubTicksCount write SetSubTicksCount default 5;
    property Title: string read FTitle write SetTitle;
    property TitleAngle: integer read FTitleAngle write SetTitleAngle default 0;
    property TitleFont:TFont read FTitleFont write SetTitleFont;
    property TickLen: integer read FTickLen write SetTickLen default 5;
    property Ticks: boolean read FTicks write SetTicks default true;
    property TicksCount: integer read FTicksCount write SetTicksCount default 6;
    property TickStep: double read FTickStep write SetTickStep;
    property Visible: boolean read FVisible write SetVisible default true;
    property LineThickness: Single read FLineThickness write SetLineThickness;
  end;

  TGPXYGraphSeries = class;
  TGPXYGraphSeriesClass = class of TGPXYGraphSeries;
  TGraphGridStyle = (gsNone, gsHorizontal, gsVertical, gsBoth);
  TZoomingData = record
    XMax: double;
    XMin: double;
    YMax: double;
    YMin: double;
    XAutoMax: boolean;
    XAutoMin: boolean;
    YAutoMax: boolean;
    YAutoMin: boolean;
  end;

  {TGPXYGraph}
  TGPXYGraph = class(TCustomControl)
  private
    {GDI+֌W}
    FGPGraphics: TGPGraphics;

    FAutoAspectRatio: boolean;
    FAspectRatioFixed: boolean;
    FBackColor: TColor;
    FXAxis: TGPGraphAxis;
    FYAxis: TGPGraphAxis;
    FGridPen: TPen;
    FGridStyle: TGraphGridStyle;
    FSubGrid: boolean;
    FSubGridPen: TPen;
    FTitleFont: TFont;
    FTitle: TStringList;
    FTitleAlignment: TAlignment;
    FPiSymbol: string;
    FWorkRect: TRect;
    FPlotArea: TRect;
    FPlotWidth: integer;
    FPlotHeight: integer;
    FSeriesList: TList;
    FMetafiling: boolean;
    FAxesFront: boolean;
    FInitialized:  boolean;
    {Zooming}
    FZooming: Boolean;
    FAnchorPt: TPoint;
    FDragPt: TPoint;
    FAllowZoom: boolean;
    FZoomed: boolean;
    FZoomingData: TZoomingData;
    {Event}
    FOnAfterDraw: TNotifyEvent;
    FOnZoom: TNotifyEvent;
    FOnUndoZoom: TNotifyEvent;
    FAspectRatio: double;
    { Catching paint events }
    FOnPaint: TNotifyEvent;
    FSubGridThickness: Single;
    FGridThickness: Single;
    FFrame: boolean;
    procedure SetAutoAspectRatio(const Value: boolean);
    procedure SetBackColor(const Value: TColor);
    procedure SetGridPen(const Value: TPen);
    procedure SetGridStyle(const Value: TGraphGridStyle);
    procedure SetPiSymbol(const Value: string);
    procedure SetSubGrid(const Value: boolean);
    procedure SetSubGridPen(const Value: TPen);
    procedure SetTitle(const Value: TStringList);
    procedure SetTitleAlignment(const Value: TAlignment);
    procedure SetTitleFont(const Value: TFont);
    procedure SetAxesFront(const Value: boolean);
    procedure SetAspectRatio(const Value: double);
    procedure SetAspectRatioFixed(const Value: boolean);
    procedure SetGridThickness(const Value: Single);
    procedure SetSubGridThickness(const Value: Single);
    procedure SetFrame(const Value: boolean);

  protected
    procedure Paint; override;

    procedure AdjustGraphArea;
    procedure ArrangeAxes; virtual;
    procedure ArrangeFrame;
    procedure DrawGraph(g: TGPGraphics);
    procedure DrawAxes(g: TGPGraphics);
    procedure DrawFrame(g: TGPGraphics);
    procedure DrawGridLines(g: TGPGraphics);
    procedure DrawBackground(g: TGPGraphics);
    procedure DrawSeries(g: TGPGraphics);
    procedure DrawTitle(g: TGPGraphics);
    procedure Error(const Msg:string);
    procedure Loaded; override;
    function CreateGraphMetafile(EMF: boolean): TMetafile;
    function CreateGraphPNGImage: TPNGImage;
    function CreateGraphGifImage: TGifImage;
    function CreateGPPen(APen: TPen; AThickness: Single): TGPPen;
    function GetSeriesCount: integer;
    function GetSeries(Index: integer): TGPXYGraphSeries;
    function IsFreeColor(Value: TColor): boolean;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure StartZooming(Pt: TPoint);
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure ZoomingTo(Pt: TPoint);
    procedure EndZooming(Pt: TPoint);
    procedure DrawZoomingRect(const Rect: TRect);
    procedure ReadState(Reader: TReader); override;
    //procedure SetClipRect(ACanvas: TCanvas; Rct: TRect);

    procedure SetPlotArea;
    procedure SetSeries(Index: integer; ASeries: TGPXYGraphSeries);
    procedure StyleChanged(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    procedure Init;
    procedure AddSeries(ASeries: TGPXYGraphSeries);
    procedure RemoveSeries(ASeries: TGPXYGraphSeries);
    procedure RemoveAllSeries;
    procedure DeleteAllSeries;
    function CreateGraphBitmap: TBitmap;
    procedure CopyToClipboardBitmap;
    procedure CopyToClipboardPngImage;
    procedure CopyToClipboardMetafile(EMF: boolean);
    procedure SaveToMetafile(const Filename: string);
    procedure SaveToMetafileEnh(const Filename: string);
    procedure SaveGraphToBitmapFile(const Filename: string);
    procedure SaveGraphToPngFile(const Filename: string);
    procedure SaveGraphToGifFile(const Filename: string);
    procedure SeriesDown(ASeries: TGPXYGraphSeries);
    procedure SeriesUp(ASeries: TGPXYGraphSeries);
    procedure ZoomRect(R: TRect);
    procedure UndoZoom;
    procedure UpdateClipLimits(ASeries: TGPXYGraphSeries);
    function SeriesIndexOf(ASeries: TGPXYGraphSeries): integer;
    function GetFreeColor: TColor;
    function GetXYPoint(X, Y: Double): TPoint;
    function MinXValue: double;
    function MaxXValue: double;
    function MinYValue: double;
    function MaxYValue: double;
    property PlotArea: TRect read FPlotArea;
    property SeriesCount: integer read GetSeriesCount;
    property Series[Index: integer]: TGPXYGraphSeries read GetSeries write SetSeries; default;
    property GPGraphics: TGPGraphics read FGPGraphics;
    property Zoomed: Boolean read FZoomed;
    { The Paint event }
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property Canvas;
  published
    property XAxis : TGPGraphAxis read FXAxis write FXAxis;
    property YAxis : TGPGraphAxis read FYAxis write FYAxis;
    property AllowZoom: boolean read FAllowZoom write FAllowZoom default True;
    property AspectRatio: double read FAspectRatio write SetAspectRatio;
    property AutoAspectRatio: boolean read FAutoAspectRatio write SetAutoAspectRatio default false;
    property AspectRatioFixed: boolean read FAspectRatioFixed write SetAspectRatioFixed default false;
    property AxesFront: boolean read FAxesFront write SetAxesFront default True;
    property BackColor: TColor read FBackColor write SetBackColor default clBtnFace;
    property Frame: boolean read FFrame write SetFrame default false;
    property GridPen:TPen read FGridPen write SetGridPen;
    property GridStyle: TGraphGridStyle read FGridStyle write SetGridStyle default gsNone;
    property GridThickness: Single read FGridThickness write SetGridThickness;
    property SubGrid: boolean read FSubGrid write SetSubGrid default false;
    property SubGridThickness: Single read FSubGridThickness write SetSubGridThickness;
    property PiSymbol: string read FPiSymbol write SetPiSymbol;
    property SubGridPen:TPen read FSubGridPen write SetSubGridPen;
    property Title:TStringList read FTitle write SetTitle;
    property TitleFont:TFont read FTitleFont write SetTitleFont;
    property TitleAlignment:TAlignment read FTitleAlignment write SetTitleAlignment default taLeftJustify;
    property OnAfterDraw: TNotifyEvent read FOnAfterDraw write FOnAfterDraw;
    property OnZoom: TNotifyEvent read FOnZoom write FOnZoom;
    property OnUndoZoom: TNotifyEvent read FOnUndoZoom write FOnUndoZoom;
    property Initialized: Boolean read FInitialized write FInitialized;
    property Align;
    property Anchors;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    property OnResize;
  end;

  TGPXYGraphSeriesAfterAdd = procedure(Sender: TGPXYGraphSeries; Index: integer) of object;

  {TXYGraphSeries}
  TGPXYGraphSeries = class(TComponent)
  private
    FParentGraph: TGPXYGraph;
    FXValues: TGraphValueList;
    FYValues: TGraphValueList;
    FSeparators: TGraphBooleanList;
    FActive: boolean;
    FOnAfterAdd: TGPXYGraphSeriesAfterAdd;
    procedure SetParentGraph(Value: TGPXYGraph);
    function GetXValue(Index: integer): double;
    procedure SetXValue(Index: integer; Value: double);
    function GetYValue(Index: integer): double;
    procedure SetYValue(Index: integer; Value: double);
    function GetSeparatorValue(Index: integer): Boolean;
    procedure SetSeparatorValue(Index: integer; Value: Boolean);
    procedure SetActive(const Value: boolean);
  protected
    FGraphXmin, FGraphXmax, FGraphYmin, FGraphYmax: double;
    FSeriesColor: TColor;
    FUpdateCount: Integer;
    FGPPen: TGPPen;
    FGPBrush: TGPBrush;
    procedure Changed; virtual;
    procedure DeviceChanged(Sender: TObject);
    function CreateGPPen(APen: TPen): TGPPen;  virtual; abstract;
    function CreateGPBrush(ABrush: TBrush): TGPBrush;  virtual; abstract;
    procedure Draw(g: TGPGraphics); virtual;
    procedure Error(const Msg:string);
    function GetIndex: integer;
    function IsSeparator(const I: Integer): boolean;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetSeriesColor(Value: TColor); virtual;
    procedure SetParentComponent(Value: TComponent); override;
    procedure UpdateValues; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure MoveTo(Dest: TGPXYGraph);
    function HasParent: Boolean; override;
    function GetParentComponent: TComponent; override;
    function DataCount: integer; virtual;
    procedure Clear; virtual;
    procedure Delete(Index: integer); virtual;
    procedure BeginUpdate;
    procedure EndUpdate;
    class function GetSeriesTitle: string; virtual; abstract;
    procedure FillSamples(Value: integer); virtual;
    function AddXY(X,Y: double): Integer; virtual;
    function AddSeparator: Integer;  virtual;
    function MinXValue: double; virtual;
    function MaxXValue: double; virtual;
    function MinYValue: double; virtual;
    function MaxYValue: double; virtual;
    property XValue[Index :integer]: double read GetXValue write SetXValue;
    property YValue[Index :integer]: double read GetYValue write SetYValue;
    property SeparatorValue[Index :integer]: Boolean read GetSeparatorValue write SetSeparatorValue;
    property XValues: TGraphValueList read FXValues;
    property YValues: TGraphValueList read FYValues;
    property Separators: TGraphBooleanList read FSeparators;
  published
    property ParentGraph: TGPXYGraph read FParentGraph write SetParentGraph;
    property SeriesColor: TColor read FSeriesColor write SetSeriesColor;
    property Active: boolean read FActive write SetActive default true;
    property OnAfterAdd: TGPXYGraphSeriesAfterAdd read FOnAfterAdd write FOnAfterAdd;

  end;{class TXYGraphSeries}

  TGPXYGraphSeriesList = array of TGPXYGraphSeries;
  function GetTextSize(g: TGPGraphics; S: string; Font: TFont): TSize;
  procedure CopyPngImageToClipboard(const png: TPngImage);

implementation

uses
// Delphi 6 UP
  RTLConsts;

procedure CopyPngImageToClipboard(const png: TPngImage);
var
  DataStream: TMemoryStream;
  Data: Pointer;
  DataHandle: THandle;
  ClipboardFormat: Integer;
begin
  if png = nil then Exit;

  ClipboardFormat := RegisterClipboardFormat('PNG');
  DataStream := TMemoryStream.Create;
  try
    png.SaveToStream(DataStream);
    DataHandle := GlobalAlloc(GMEM_DDESHARE, DataStream.Size);
    Data := GlobalLock(DataHandle);
    try
      DataStream.Position := 0;
      DataStream.ReadBuffer(Data^, DataStream.Size);
    finally
      GlobalUnlock(DataHandle);
    end;
    Clipboard.SetAsHandle(ClipboardFormat, DataHandle);
  finally
    DataStream.Free;
  end;
end;

function TGraphValueList.Add(Value: double): Integer;
begin
  Result := inherited Add(Value);
  PutValue(Result, Value);
end;

{TGraphValueList}
procedure TGraphValueList.Assign(Source: TGraphValueList);
var
  I: Integer;
begin
  Clear;
  for I := 0 to Source.Count - 1 do
    Add(Source.Items[I])
end;

procedure TGraphValueList.PutValue(Index: Integer; Value: double);
begin
  if Count=1 then
  begin
    FMaxValue:=Value;
    FMinValue:=Value;
    FMinValueIndex:=Index;
    FMaxValueIndex:=Index;
  end
  else if (Index = FMinValueIndex) or (Index = FMinValueIndex) then
    RecalcMinMax
  else
  begin
    if Value>FMaxValue then
    begin
      FMaxValue:=Value;
      FMaxValueIndex:=Index;
    end;
    if Value<FMinValue then
    begin
      FMinValue:=Value;
      FMinValueIndex:=Index;
    end;
  end;

end;

procedure TGraphValueList.RecalcMinMax;
var
  i: integer;
begin
  if Count=0 then Exit;
  FMaxValue:=List[0];
  FMinValue:=FMaxValue;
  for i:=1 to Count-1 do
  begin
    if Items[i]>FMaxValue then
    begin
      FMaxValue:=Items[i];
      FMaxValueIndex:=i;
    end;
    if Items[i]<FMinValue then
    begin
      FMinValue:=Items[i];
      FMinValueIndex:=i;
    end;
  end;
end;

{ TGraphPointList }
function TGraphPointList.AddXY(X, Y: integer): Integer;
begin
  Add(Point(X, Y));
end;

function GetTextSize(g: TGPGraphics; S: string; Font: TFont): TSize;
var
  GPFontFamily: TGPFontFamily;
  SizeRectF: TGPRectF;
  GPFont: TGPFont;
  GPFontSize: Single;
  APoint: TGPPointF;
begin
  Result.cx := 0;
  Result.cy := 0;
  if S = '' then Exit;

  GPFontFamily := TGPFontfamily.Create(Font.Name);
  GPFontSize := Font.Size;
  GPFont := TGPFont.Create(GPFontFamily, GPFontSize, FontStyleRegular, UnitPixel);

  APoint :=  MakePoint(0.0, 0.0);
  try
    g.MeasureString(S, -1, GPFont, APoint, SizeRectF);
  finally
    GPFontFamily.Free;
    GPFont.Free;
  end;
  Result.Width := Trunc(SizeRectF.Width);
  Result.Height := Trunc(SizeRectF.Height);
end;

{ TGraphAxis }
constructor TGPGraphAxis.Create(ParentGraph: TGPXYGraph);
begin
  inherited Create;
  FParentGraph:=ParentGraph;
  FAxisPen:=TPen.Create;
  FLineThickness := 1.0;
  FAxisPen.OnChange:=AxisChanged;
  FLabelFont:=TFont.Create;
  with FLabelFont do begin
    Name:='Arial';
    Size:=15;
  end;
  FLabelFont.OnChange:=AxisChanged;
  FTitleFont:=TFont.Create;
  FTitleFont.Assign(FLabelFont);
  FTitleFont.OnChange:=AxisChanged;
  FVisible:=true;
  FOrientation:=aoHorizontal;
  FAxisBounds:=ParentGraph.BoundsRect;
//  FAxisPos:=FAxisBounds.Left;

  FTickDataList:=TGraphValueList.Create;
  FSubTickDataList:=TGraphValueList.Create;
  FLabelValueList:=TGraphValueList.Create;
  Init;
end;

function TGPGraphAxis.CreateGPPen(APen: TPen): TGPPen;
var
  Index: Integer;
begin
  //Pen.Width 1ȊȌꍇ́AÂt@CłB
  //LineTfichness 𒲐B
  if APen.Width > 1 then
    FLineThickness := APen.Width;
  Result := TGPPen.Create(ColorRefToARGB(APen.Color), FLineThickness);
  if APen.Style = psClear then
    Result.SetColor(MakeColor(0, 0, 0, 0));  //Ȑ
  if APen.Style < psClear then
     Index := Integer(APen.Style);
  Result.SetDashStyle(DashStyle(Index));
end;

procedure TGPGraphAxis.Init;
begin
  FVisible:=true;
  FAxisBounds:=ParentGraph.BoundsRect;
//  FAxisPos:=FAxisBounds.Left;
  with FLabelFont do begin
    Name:='Arial';
    Size:=15;
  end;
  with FTitleFont do begin
    Name:='Arial';
    Size:=15;
  end;
  with  FAxisPen do begin
    Width := 1;
    Style := psSolid;
    Color := clBlack;
  end;
  FAxisScale := asNormal;
  FMin:=0;
  FMax:=10;
  FRange:=FMax-FMin;
  FLogRange:=1;
  FAutoMin:=true;
  FAutoMax:=true;
  FAutoStep:=true;
  FAutoOrigin:=true;
  FTicks:=true;
  FLabels:=true;
  FAutoDecimal:=true;
  FExtraLen:=8;
  FInverted:=false;

  FTickStep:=2;
  FTicksCount:= 6;
  FCurrTicksCount:= 6;
  FSubTicksCount:=5;
  FTickLen:=5;
  FSubTickLen:=3;
  FLabelDecimal:=0;
  FLabelSpace:=4;
  FLabelCount:=5;
  FTitle := '';
  FTitleAngle := 0;
  FAxisTitleLoc := tlHead;
  UpdateScale;
end;


destructor TGPGraphAxis.Destroy;
begin
  FLabelValueList.Free;
  FSubTickDataList.Free;
  FTickDataList.Free;
  FTitleFont.Free;
  FLabelFont.Free;
  FAxisPen.Free;
  inherited Destroy;
end;

procedure TGPGraphAxis.SetAxisBounds(const Value: TRect);
begin
  FAxisBounds:=Value;
  AdjustSize;
end;

procedure TGPGraphAxis.SetAxisPen(const Value: TPen);
begin
  FAxisPen.Assign(Value);
end;

procedure TGPGraphAxis.SetLabelFont(const Value: TFont);
begin
  FLabelFont.Assign(Value);
end;

procedure TGPGraphAxis.SetTitleFont(const Value: TFont);
begin
  FTitleFont.Assign(Value);
end;

procedure TGPGraphAxis.SetTitle(const Value: string);
begin
  if FTitle<>Value then
  begin
    FTitle:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetTitleAngle(const Value: integer);
var
  Temp: Integer;
begin
  Temp := 90*((((Value Mod 360) +360) mod 360) div 90);
  if Temp <> FTitleAngle then
  begin
    FTitleAngle := Temp;
    AxisChanged(Self);
  end;
end;

function TGPGraphAxis.TitleSize: TSize;
var
Temp: TSize;
begin
  Result.Width := 0; Result.Height := 0;
  if Title = '' then Exit;
  if Parentgraph.GPGraphics = nil then Exit;

  Result := GetTextSize(Parentgraph.GPGraphics, Title, TitleFont);
  if (FTitleAngle = 90) or (FTitleAngle = 270)  then
  begin
    Temp := Result;
    Result.Width := Temp.Height;
    Result.Height := Temp.Width;
  end;
end;

function TGPGraphAxis.TitleHeight: integer;
begin
  Result := TitleSize.Height;
end;

function TGPGraphAxis.TitleWidth: integer;
begin
  Result := TitleSize.Width;
end;

procedure TGPGraphAxis.SetOrientation(const Value: TAxisOrientation);
begin
  if Value <> FOrientation then
  begin
    FOrientation:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAutoStep(const Value: boolean);
begin
  if FAutoStep <>Value then
  begin
    FAutoStep:=Value;
    if FAutoStep and not (csLoading in ParentGraph.ComponentState)then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAutoOrigin(const Value: boolean);
begin
  if Value <> FAutoOrigin then
  begin
    FAutoOrigin:=Value;
    if FAutoStep and not (csLoading in ParentGraph.ComponentState)then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetVisible(const Value: boolean);
begin
  if Value <> FVisible then
  begin
    FVisible:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetTicks(const Value: boolean);
begin
  if Value <> FTicks then
  begin
    FTicks:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetLabels(const Value: boolean);
begin
  if Value <> FLabels then
  begin
    FLabels:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetLabelsOnAxis(const Value: boolean);
begin
  if Value <> FLabelsOnAxis then
  begin
    FLabelsOnAxis:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAutoDecimal(const Value: boolean);
begin
  if Value <> FAutoDecimal then
  begin
    FAutoDecimal:=Value;
    if FAutoDecimal then
      FLabelDecimal := 0;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAutoMin(const Value: boolean);
begin
  if FAutoMin <> Value then
  begin
    FAutoMin := Value;
    if FAutoMin and not (csLoading in ParentGraph.ComponentState)then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAutoMax(const Value: boolean);
begin
  if FAutoMax <> Value then
  begin
    FAutoMax := Value;
    if FAutoMax and not (csLoading in ParentGraph.ComponentState) then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetMax(const Value: double);
var
  Changed: boolean;
begin
  Changed := false;
  if Value <> FMax then
  begin
    FMax := Value;
    Changed := true;
  end;
  {ǂݍ݂̎ɂ͕pNȂ}
  if csLoading in ParentGraph.ComponentState then Exit;
  FAutoMax:=false;
  if Changed then
  begin
    SetAxisRange(FMin, FMax);
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetMin(const Value: double);
var
  Changed: boolean;
begin
  Changed := false;
  if Value <> FMin then
  begin
    FMin := Value;
    Changed := true;
  end;
  if csLoading in ParentGraph.ComponentState then Exit;
  FAutoMin:=false;
  if Changed then
  begin
    SetAxisRange(FMin, FMax);
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetMinMax(const AMin, AMax: double);
begin
  FMin := AMin;
  FAutoMin := False;
  FMax := AMax;
  FAutoMax := False;
  AxisChanged(Self);
end;

procedure TGPGraphAxis.SetAxisRange(const AMin, AMax: double);
begin
  if AMin>=AMax then
    Error('͈̔͂𒴂Ă܂');
  if (FAxisScale=asLog) and (AMin<=0) then
    Error('ΐڐł͎͈̔͂̓[傫ȂƂ܂');
  FMin:=AMin;
  FMax:=AMax;
  FRange:=AMax-AMin;
  if FRange<MinimumRange then
  begin
    FRange:=MinimumRange;
    FMin:=AMin;
    FMax:=FMin+FRange;
  end;
  if FOrigin > FMax then
    FOrigin := FMax;
  if FOrigin < FMin then
    FOrigin := FMin;
  if FMin>0 then
    FLogRange:=Log10(FMax/FMin);
end;

procedure TGPGraphAxis.SetTickStep(const Value: double);
begin
  if csLoading in ParentGraph.ComponentState then
  begin
    FTickStep:=Value;
    Exit;
  end;

  FAutoStep:=false;
  if (Value <> FTickStep) and CheckTickStep(Value) then
  begin
    FTickStep:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetTicksCount(const Value: integer);
begin
  if (Value > 0) and (Value <> FTicksCount) and (Value < MaxTicksCount) then
  begin
    FTicksCount:=Value;
    FCurrTicksCount:= Value;
    if not (csLoading in ParentGraph.ComponentState) then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetSubTicksCount(const Value: integer);
begin
  if (Value > 0) and (Value <> FSubTicksCount) and (Value<MaxSubTicksCount) then
  begin
    if FAxisScale=asLog then Exit;
    FSubTicksCount:=Value;
    if not (csLoading in ParentGraph.ComponentState) then
      AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetTickLen(const Value: integer);
begin
  if Value <> FTickLen then
  begin
    FTickLen:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetSubTickLen(const Value: integer);
begin
  if Value <> FSubTickLen then
  begin
    FSubTickLen:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetOrigin(const Value: double);
begin
  if Value <> FOrigin then
  begin
    if csLoading in ParentGraph.ComponentState then
    begin
      FOrigin:=Value;
      Exit;
    end;
    if Value < FMin then
      Error('̌_̒lŏlȉł');
    if Value > FMax then
      Error('̌_̒lől𒴂Ă܂');
    FOrigin:=Value;
    AutoOrigin:=false;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAxisScale(const Value: TAxisScale);
begin
  if Value<>FAxisScale then
  begin
    if csLoading in ParentGraph.ComponentState then
    begin
      FAxisScale:=Value;
      Exit;
    end;
    if (Value=asLog) and ((FMin<=0) or (GetSeriesMin<=0)) then
      Error('ŏl[ȉ̂߁Aΐڐɂł܂');
    FAxisScale:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetAxisTitleLoc(const Value: TAxisTitleLoc);
begin
  if Value <> FAxisTitleLoc then
  begin
    FAxisTitleLoc:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetExtraLen(const Value: integer);
begin
  if Value <> FExtraLen then
  begin
    FExtraLen:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetInverted(const Value: boolean);
begin
  if Value <> FInverted then
  begin
    FInverted:=Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetLabelDecimal(const Value: integer);
begin
  if Value < 0 then
    Error('̒l͖ł');
  if (Value < 10) and (Value <> FLabelDecimal) then
  begin
    FLabelDecimal:=Value;
    FAutoDecimal:=false;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.SetLabelSpace(const Value: integer);
begin
  if Value <> FLabelSpace then
  begin
    FLabelSpace:=Value;
    AxisChanged(Self);
  end;
end;

function TGPGraphAxis.PiSymbol: string;
begin
  Result := ParentGraph.PiSymbol;
end;

function TGPGraphAxis.GetSeriesMin :double;
begin
  if FOrientation = aoHorizontal then
    Result := ParentGraph.MinXValue
  else
    Result := ParentGraph.MinYValue;
end;

function TGPGraphAxis.GetSeriesMax :double;
begin
  if FOrientation = aoHorizontal then
    Result := ParentGraph.MaxXValue
  else
    Result := ParentGraph.MaxYValue;
end;

function TGPGraphAxis.GetOriginPos: integer;
begin
  Result:=GetValuePos(FOrigin);
end;

function TGPGraphAxis.GetTitleHeight(Loc: TAxisTitleLoc) :integer;
begin
  if (FAxisTitleLoc=Loc) and FVisible  then
    Result:=TitleHeight + LabelSpace
  else
    Result:=0;
end;

function TGPGraphAxis.GetTitleWidth(Loc: TAxisTitleLoc) :integer;
begin
  if (FAxisTitleLoc=Loc) and FVisible then
    Result:=TitleWidth
  else
    Result:=0;
end;

function TGPGraphAxis.GetAxisHeight: integer;
begin
  if FOrientation = aoHorizontal then
    Result:=GetMaxLabelsHeight+FLabelSpace+GetTitleHeight(tlSide)
  else
    Result:=FAxisBounds.Right - FAxisBounds.Left;
end;

function TGPGraphAxis.GetAxisWidth: integer;
begin
  if FOrientation = aoHorizontal then
    Result:=FAxisBounds.Right - FAxisBounds.Left
  else
    Result:=GetMaxLabelsWidth + FLabelSpace+GetTitleWidth(tlSide);
  if 2*Result<GetTitleWidth(tlHead)+GetTitleWidth(tlFoot) then
    Result:=(GetTitleWidth(tlHead)+GetTitleWidth(tlFoot)) div 2;
end;

procedure TGPGraphAxis.AdjustSize;
{
  ̔zu߂
    ̘g AxisBounds
    tHg̃TCY
    őlэŏlA
    x̃tH[}bg
  ɈˑB
}
var
  Margin, AxisSize, LenToNext: integer;
  Temp, FirstValue, LastValue: double;
begin
  if FOrientation = aoHorizontal then
  begin
    with FAxisBounds do
    begin
      AxisSize := Right - Left;
      Margin := FExtraLen + GetTitleWidth(tlHead);
      {3̈ȏ鎲̃^CǵAz肵Ȃ}
      if Margin > AxisSize div 3 then
        Margin := AxisSize div 3;
      FMaxPos:=Right - Margin;
      Margin := FExtraLen + GetTitleWidth(tlFoot);
      if Margin > AxisSize div 3 then
        Margin := AxisSize div 3;
      FMinPos := Left + Margin;
    end;
    {PlotSize ̎blvZ}
    FPlotSize:=FMaxPos - FMinPos;
    {܂x҂鐔ɂāAXebvlXV}
    FCurrTicksCount := FTicksCount;
    UpdateScale;
    {ׂ荇ẙԊuől1.5{}
    LenToNext := 3 * GetMaxLabelsWidth div 2;
    while FAutoStep and (FCurrTicksCount > 1)
      and (LenToNext * FCurrTicksCount > FPlotSize) do
    begin
      Dec(FCurrTicksCount);
      UpdateScale;
    end;
    {xl }
    if FLabelValueList.Count > 0 then
    begin
      FirstValue := FLabelValueList.MinValue;
      LastValue := FLabelValueList.MaxValue;
      if Inverted then
      begin
       Temp := FirstValue;
       FirstValue := LastValue;
       LastValue := Temp;
     end;
      Margin := GetLabelWidth(LastValue) div 2 - FMaxPos + GetValuePos(LastValue);
      if Margin > 0 then
      begin
        if Margin > FPlotSize div 3 then
          Margin := FPlotSize div 3;
        FMaxPos := FMaxPos - Margin;
      end;
      Margin := GetLabelWidth(FirstValue) div 2 - GetValueSize(FirstValue);
      if Margin > 0 then
      begin
        if Margin > FPlotSize div 3 then
          Margin := FPlotSize div 3;
        FMinPos := FMinPos + Margin;
      end;
    end;
    FStartPos := FMinPos - FExtraLen;
    FEndPos := FMaxPos + FExtraLen - 1; {trable shooting}
    FPlotSize := FMaxPos - FMinPos;
  end
  else
  begin
    with FAxisBounds do
    begin
      AxisSize := Bottom - Top;
      Margin := FExtraLen + GetTitleHeight(tlFoot);
      if Margin > AxisSize div 3 then
        Margin := AxisSize div 3;
      FMinPos := Bottom - Margin;
      Margin := FExtraLen + GetTitleHeight(tlHead);
      if Margin > AxisSize div 3 then
        Margin := AxisSize div 3;
      FMaxPos := Top + Margin;
    end;
    {PlotSize ̎blvZ}
    FPlotSize := FMinPos - FMaxPos;
    {܂x҂鐔ɂāAXebvlXV}
    FCurrTicksCount := FTicksCount;
    UpdateScale;
    {ׂ荇ẙԊuől1.5{}
    LenToNext := 3 * GetMaxLabelsHeight div 2;
    while FAutoStep and (FCurrTicksCount > 1)
      and (LenToNext * FCurrTicksCount > FPlotSize) do
    begin
      Dec(FCurrTicksCount);
      UpdateScale;
    end;
    {xl }
    if FLabelValueList.Count > 0 then
    begin
      FirstValue := FLabelValueList.MinValue;
      LastValue := FLabelValueList.MaxValue;
      if Inverted then
      begin
       Temp := FirstValue;
       FirstValue := LastValue;
       LastValue := Temp;
     end;
      Margin := GetLabelHeight(LastValue) div 2 + FMaxPos - GetValuePos(LastValue);
      if Margin > 0 then
      begin
        if Margin > FPlotSize div 3 then
          Margin := FPlotSize div 3;
        FMaxPos := FMaxPos + Margin;
      end;
      Margin := GetLabelHeight(FirstValue) div 2 - GetValueSize(FirstValue);
      if Margin > 0 then
      begin
        if Margin > FPlotSize div 3 then
          Margin := FPlotSize div 3;
        FMinPos := FMinPos - Margin;
      end;
    end;
    FStartPos := FMinPos + FExtraLen;
    FEndPos := FMaxPos - FExtraLen;
    FPlotSize := FMinPos - FMaxPos;
  end;
end;

function TGPGraphAxis.CalcOriginValue(AMin, AMax, AStep: double): double;
begin
  Result:=AMin;
  if AStep=0 then Exit;
  if (AMax>=0) and (AMin<=0) then
    Result:=0
  else if (AMin > 0) and FAutoMin then
    Result:=AStep*Floor(AMin/AStep + epsilon)
  else if (AMax < 0) and FAutoMax then
    Result:=AStep*Ceil(AMax/AStep + epsilon);
end;

procedure TGPGraphAxis.AddTickData(Value: double);
var
  I: Integer;
begin
  if (FMin <= Value) and (Value <= FMax) then
  begin
    if Iszero(Value){(abs(Value) < 1E-14)} then
      Value := 0;
    for I := 0 to FTickDataList.Count - 1 do
      if FTickDataList.Items[I] = Value then Exit;
    FTickDataList.Add(Value);
  end;
end;

procedure TGPGraphAxis.AddSubTickData(Value: double);
var
  I: Integer;
begin
  if (FMin <= Value) and (Value <= FMax) then
  begin
    for I := 0 to FSubTickDataList.Count - 1 do
      if FSubTickDataList.Items[I] = Value then Exit;
    FSubTickDataList.Add(Value);
  end;
end;

function TGPGraphAxis.CheckTickStep(Value: double): boolean;
begin
  Result:=true;
  if Value <= 0 then
    Error('ڐԊu̓[ȉɐݒł܂')
  else if Value>FRange then
    Error('ڐԊu͈̔͂𒴂Ă܂')
  else if FRange/Value>MaxTicksCount then
    Error('ڐԊu܂');
end;

function TGPGraphAxis.GetValueText(Value: double):string;
var
  Format: TFloatFormat;
  Precision: Integer;
begin
  if FAutoDecimal then
    Format:=ffGeneral
  else
    Format:=ffFixed;
  case FAxisScale of
  asPi:
    Value:=Value/pi;
  asPercent:
    Value:=Value/100;
  end;
  if abs(Value)<epsilon then
    Value:=0;
  Precision := 14 + FLabelDecimal;
  if Precision > 14 then
    Precision := 14;
  Result:=FloatToStrF(Value, Format, Precision, FLabelDecimal);
  case FAxisScale of
  asPi:
    begin
      if Value=0 then
        Result:='0'
      else if (Result='1') and (PiSymbol<>'') then
        Result:=PiSymbol
      else if (Result='-1') and (PiSymbol<>'') then
        Result:='-'+PiSymbol
      else
        Result:=Result+PiSymbol
    end;
  asPercent:
    Result:=Result+'%';
  end;
end;

function TGPGraphAxis.GetLabelHeight(Value: double):integer;
var
  Size: TSize;
begin
  //ParentGraph.Canvas.Font := LabelFont;
  //Result := ParentGraph.Canvas.TextHeight(GetValueText(Value));
  //{́AGDI+ ̎葱gׂłB
  Size := GetTextSize(Parentgraph.GPGraphics, GetValueText(Value), LabelFont);
  Result := Size.cy;
end;

function TGPGraphAxis.GetLabelWidth(Value: double):integer;
var
  Size: TSize;
begin
  //ParentGraph.Canvas.Font := LabelFont;
  //Result := ParentGraph.Canvas.TextWidth(GetValueText(Value));
  Size := GetTextSize(Parentgraph.GPGraphics, GetValueText(Value), LabelFont);
  Result := Size.cx;
end;

function TGPGraphAxis.GetMaxLabelsWidth:integer;
var
  I, Index, W, Len: integer;
  S: String;
begin
  Len := 0;
  Index := 0;
  for I:=0 to FLabelValueList.Count-1 do
  begin
    S := FloatToStr(FLabelValueList[I]);
    W := S.Length;
    if W > Len then
    begin
      Len := W; Index := I;
    end;
  end;
  Result := GetLabelWidth(FLabelValueList[Index]);
end;

function TGPGraphAxis.GetMaxLabelsHeight:integer;
begin
  Result := GetLabelHeight(FLabelValueList[0]);
end;

function TGPGraphAxis.GetValueSize(Value: double):integer;
var
  V: double;
begin
  if FAxisScale = asLog then
  begin
    if Value<=0 then
      Error('ΐڐŒl[ȉł');
    if FInverted then
      V:=Log10(FMax/Value)/FLogRange
    else
      V:=Log10(Value/FMin)/FLogRange;
  end
  else
  begin
    if FInverted then
      V:=(FMax-Value)/FRange
    else
      V:=(Value-FMin)/FRange;
  end;

  if V>MaxInt then                 //?
    V:=2;
  if V<-MaxInt then
    V:=-2;
  Result:=Round(FPlotSize * V);
end;

function TGPGraphAxis.GetValuePos(Value: double):integer;
var
  Len: integer;
begin
  Len:=GetValueSize(Value);
  if FOrientation = aoHorizontal then
    Result := FMinPos + Len
  else
    Result := FMinPos - Len;
end;

function TGPGraphAxis.GetPosValue(Pos: integer): double;
var
  Len: integer;
  Delta: double;
begin
  if FOrientation = aoHorizontal then
    Len := Pos - FMinPos
  else
    Len := FMinPos - Pos;
  if (Len < 0 ) or (FPlotSize < Len) then
    Error('͈̔͂𒴂Ă܂B');

  Delta := Len / FPlotSize;
  if FAxisScale = asLog then
  begin
    if not Inverted then
      Result := FMin + Power(10, Delta * FLogRange)
    else
      Result := FMax - Power(10, Delta * FLogRange);
  end
  else
  begin
    if not FInverted then
      Result := Min + FRange * Delta
    else
      Result := Max - FRange * Delta;
  end;
end;

procedure TGPGraphAxis.AxisChanged(Sender: TObject);
begin
  {ǂݍ݂̎ɂ͕pNȂ}
  if csLoading in ParentGraph.ComponentState then Exit;

  UpdateScale;
  ParentGraph.Invalidate;
end;

procedure TGPGraphAxis.Error(const Msg:string);
begin
  raise EXYGraphAxisError.Create(Msg);
end;

function TGPGraphAxis.CalcTickStep(Range: double):double;
var
  Temp, Step, Scale: double;
begin
  if FAxisScale=asPi then
    Temp:=Range/pi
  else
    Temp:=Range;

  Step := Temp / FCurrTicksCount;
  Scale := IntPower(10, Floor(Log10(Step)));
  Step := Step / Scale;
  if Step < 1 then
  begin
    Step := 10 * Step;
    Scale := Scale / 10;
  end
  else if Step >= 10 then
  begin
    Step := Step / 10;
    Scale := 10 * Scale;
  end;
  if Step <= 1 then
    Result := Scale
  else if Step <= 2 then
    Result := 2 * Scale
  else if (FAxisScale <> asPi) and (Step <= 2.5) then
    Result := 2.5 * Scale
  else if Step <= 5 then
    Result := 5 * Scale
  else
    Result := 10*Scale;
  if FAxisScale=asPi then
    Result:=pi*Result;
end;

procedure TGPGraphAxis.SetTickValues;
var
  i : integer;
  SubTickStep, Major, Minor, epsl: double;
begin
  SubTickStep := FTickStep / FSubTicksCount;
  FTickDataList.Clear;
  FSubTickDataList.Clear;
  epsl:=0.001 * FTickStep;

  Major := Ceil(FMin / FTickStep -0.05) * FTickStep;
  if not FAutoOrigin then
    Major := FOrigin;

  while Major > FMin + epsl do
    Major := Major - FTickStep;

  while Major < FMax + epsl do
  begin
    if (Major > FMin - epsl) and (Major < FMax + epsl) then
      AddTickData(Major);
    for i := 1 to FSubTicksCount-1 do
    begin
      Minor := Major + i * SubTickStep;
      if (Minor > FMin) and (Minor < FMax) then
        AddSubTickData(Minor);
    end;
    Major := Major + FTickStep;
  end;
  SetLabelValues;
end;

const
  LogRef : array[0..14] of double =
             (1, 5, 2, 3, 1.5, 7, 4, 6, 1.2, 8, 9, 1.3, 2.5, 1.1, 1.4);
  LogRefLen = 14;

procedure TGPGraphAxis.SetLogTickValues;
var
  I, J, N, Sub, iMin, iMax:integer;
  Temp, Scale: double;
begin
  FTickDataList.Clear;
  FSubTickDataList.Clear;
  iMin := Floor(Log10(FMin));
  iMax := Floor(Log10(FMax));
  if iMax = iMin then
    Inc(iMax);
  N := Floor(FCurrTicksCount/(iMax - iMin));
  if N > 0 then
  begin
    if N > LogRefLen then N := LogRefLen;
    for I := iMin to iMax do
    begin
      Scale := IntPower(10, I);
      for J := 0 to N - 1 do
        AddTickData(LogRef[J] * Scale);
      Sub := N + FSubTicksCount;
      if Sub > LogRefLen then Sub := LogRefLen;
      for J := N to Sub do
        AddSubTickData(LogRef[J] * Scale / 10);
    end;
  end
  else
  begin
    N := Ceil((iMax - iMin)/FCurrTicksCount);
    I := iMin;
    while I <= iMax do
    begin
      Scale := IntPower(10, I);
      AddTickData(Scale);
      J := N - 1;
      Temp := Scale / 10;
      while J > 0  do
      begin
        AddSubTickData(Temp);
        Temp := Temp / 10;
        Dec(J);
      end;
      Sub := FSubTicksCount - 1;
      if Sub > LogRefLen then Sub := LogRefLen;
      for J := 1 to Sub do
        AddSubTickData(LogRef[J] * Scale / 10);
      Inc(I, N);
    end;
  end;
  FTickDataList.Sort;
  FSubTickDataList.Sort;
  SetLabelValues;
end;

procedure TGPGraphAxis.UpdateScale;
var
  Temp, TempMin, TempMax, TempStep: double;
begin
  if csDesigning in ParentGraph.ComponentState then
  begin
    TempMax := FMax;
    TempMin := FMin;
  end
  else
  begin
    if FAutoMax and (ParentGraph.SeriesCount > 0) then
      TempMax:= GetSeriesMax
    else
      TempMax := FMax;
    if FAutoMin and (ParentGraph.SeriesCount > 0) then
      TempMin := GetSeriesMin
    else
      TempMin := FMin;
    {trable shooting}
    if (TempMax - TempMin) < MinimumRange then
    begin
      TempMin := TempMin - 1;
      TempMax := TempMax + 1;
    end;
  end;

  if FAxisScale<>asLog then
  begin
    if FAutoStep then
      TempStep := CalcTickStep(TempMax-TempMin)
    else
      TempStep := FTickStep;
    if FAutoOrigin then
      FOrigin := CalcOriginValue(TempMin, TempMax, TempStep);
    if FAutoMax then
    begin
      if TempMax >= FOrigin then
        Temp := FOrigin + TempStep * Floor((TempMax - FOrigin)/TempStep)
      else
        Temp := FOrigin;
      while Temp < TempMax do
        Temp := Temp + TempStep/10;
      FMax := Temp;
    end;
    if FAutoMin then
    begin
      if FOrigin >= TempMin then
        Temp := FOrigin - TempStep * Floor((FOrigin - TempMin)/TempStep)
      else
        Temp := FOrigin;
      while Temp > TempMin do
        Temp := Temp - TempStep/10;
      FMin := Temp;
    end;
    SetAxisRange(FMin,FMax);

    if  FAutoStep then
      FTickStep := CalcTickStep(FMax-FMin);

    SetTickValues;
  end
  else {if FAxisScale = asLog then}
  begin
    if FAutoMax then
      FMax := IntPower(10, Ceil(Log10(TempMax)));
    if FAutoMin then
      FMin := IntPower(10, Floor(Log10(TempMin)));
    if FAutoOrigin then
      FOrigin := FMin;
    SetAxisRange(FMin,FMax);
    SetLogTickValues;
  end;
end;

procedure TGPGraphAxis.SetLabelValues;
var
  i: integer;
begin
  FLabelValueList.Clear;
  for i:=0 to FTickDataList.Count-1 do
    FLabelValueList.Add(FTickDataList[i]);
end;

procedure TGPGraphAxis.SetLineThickness(const Value: Single);
begin
  if FLineThickness <> Value then
  begin
    FLineThickness := Value;
    AxisChanged(Self);
  end;
end;

procedure TGPGraphAxis.DrawTicksLines(List: TGraphValueList; Pos, Len: integer);
var
  i, ValuePos: integer;
  GPPen: TGPPen;
begin
  //ParentGraph.D2DCanvas.Pen.Assign(FAxisPen);
  if ParentGraph.FGPGraphics = nil then
    Exit;

  GPPen := CreateGPPen(AxisPen);
  try
    for i:=0 to List.Count-1 do
    begin
      ValuePos := GetValuePos(List[i]);
      if FOrientation = aoVertical then
        ParentGraph.FGPGraphics.DrawLine(GPPen, Pos, ValuePos, Pos + Len, ValuePos)
      else
        ParentGraph.FGPGraphics.DrawLine(GPPen, ValuePos, Pos, ValuePos, Pos - Len);
    end;
  finally
    GPPen.Free;
  end;
end;

procedure TGPGraphAxis.DrawAxisTicks(Pos: Integer; Reverse: Boolean);
var
  Len: Integer;
begin
  if not FTicks or not FVisible then Exit;
  Len := FTickLen;
  if Reverse then
    Len := -1 * Len;
  DrawTicksLines(FTickDataList, Pos, Len);
  Len := FSubTickLen;
  if Reverse then
    Len := -1 * Len;
  DrawTicksLines(FSubTickDataList, Pos, Len);
end;

procedure TGPGraphAxis.DrawAxisLine(Pos: Integer);
var
  GPen: TGPPen;
begin
  if not FVisible then exit;
  if ParentGraph.FGPGraphics = nil then
    Exit;
  GPen := CreateGPPen(AxisPen);
  try
    if FOrientation = aoVertical then
      ParentGraph.GPGraphics.DrawLine(GPen, Pos, FStartPos, Pos, FEndPos)
    else
      ParentGraph.GPGraphics.DrawLine(GPen, FStartPos, Pos, FEndPos, Pos);
  finally
    GPen.Free;
  end;
end;

procedure TGPGraphAxis.PutLabel(AFont: TGPFont; ABrush: TGPBrush;
  Pos: Integer; Value: double);
var
  X, Y, P: integer;
  S: String;
begin
  if not FVisible then exit;
  if (not FLabelsOnAxis) and (abs(Value-FOrigin)<epsilon) then
    Exit;
  P:=GetValuePos(Value);
  if FOrientation = aoVertical then
  begin
    X:=Pos - FLabelSpace - GetLabelWidth(Value) + 1;
    Y:=P + 1 - GetLabelHeight(Value) div 2;
  end
  else
  begin
    X:=P - GetLabelWidth(Value) div 2;
    Y:=Pos + FLabelSpace;
  end;

  S := GetValueText(Value);
  //Pt := MakePoint(X+0.5, Y+0.5);
  ParentGraph.FGPGraphics.DrawString(S, -1, AFont, MakePoint(X+0.5, Y+0.5), ABrush);
end;

procedure TGPGraphAxis.DrawAxisLabels(Pos: Integer);
var
  i: integer;
  GPFontFamily : TGPFontFamily;
  GPFontSize: Single;
  GPFont: TGPFont;
  GPBrush: TGPSolidBrush;
begin
  GPFontFamily := TGPFontfamily.Create(LabelFont.Name);
  GPFontSize := LabelFont.Size;
  GPFont := TGPFont.Create(GPFontFamily, GPFontSize, FontStyleRegular, UnitPixel);
  GPBrush := TGPSolidBrush.Create(ColorRefToARGB(LabelFont.Color));
  try
    for i:=0 to FLabelValueList.Count-1 do
      PutLabel(GPFont, GPBrush, Pos, FLabelValueList[i]);
  finally
    GPFontFamily.Free;
    GPFont.Free;
    GPBrush.Free;
  end;
end;

procedure TGPGraphAxis.DrawTitle(Pos: Integer);
var
  TW, TH: integer;
  GPFontSize, X, Y, dX, dY, Angle: Single;
  GPFont: TGPFont;
  GPBrush: TGPSolidBrush;
begin
  if Title = '' then  Exit;

  TW := TitleWidth;
  TH := TitleHeight;
  //Angle = 0

  Angle := - FTitleAngle;
  dX := 0; dY := 0;

  if FOrientation = aoVertical then
  begin
    case FTitleAngle of
        0: begin dX := - TW div 2; dY := 0; end;
       90: begin dX := - TH; dY := - TW div 2; end;
      180: begin dX := - TW div 2; dY := - TH; end;
      270: begin dX := 0; dY := - TW div 2; end;
    end;
    case FAxisTitleLoc of
    tlHead:
      begin
        X:=Pos;
        Y:=FAxisBounds.Top;
      end;
    tlFoot:
      begin
        X:=Pos;
        Y:=FAxisBounds.Bottom - TH;
      end;
    tlSide:
      begin
        Y:=FMinPos - (2*FPlotSize div 3);
        X:=Pos - 2*AxisWidth div 3;
      end;
    else//tlFixed:
      begin
        Y:=FMinPos - (2*FPlotSize div 3);
        X := TW div 2;
      end;

    end;{of case}
  end
  else {Horizontal Axis}
  begin
    case FTitleAngle of
        0: begin dX := - TW; dY := - TH div 2; end;
       90: begin dX := - TH div 2; dY := - TW; end;
      180: begin dX := 0; dY := - TH div 2; end;
      270: begin dX := - TW; dY := 0; end;
    end;
    case FAxisTitleLoc of
    tlHead:
      begin
        X:=FAxisBounds.Right;
        Y:=Pos;
      end;
    tlFoot:
      begin
        X:=FStartPos;
        Y:=Pos;
      end;
    tlSide:
      begin
        X:=FStartPos + (2*FPlotSize div 3);
        Y:=Pos + 2*AxisHeight div 3;
      end;
    else//tlFixed:
      begin
        X:=FStartPos + (2*FPlotSize div 3) - TW div 2;
        Y:=FAxisBounds.Bottom + 2*TH div 3 ;
      end;

    end;{of case}
  end;

  ParentGraph.GPGraphics.TranslateTransform(X, Y);
  ParentGraph.GPGraphics.RotateTransform(angle);
  //ParentGraph.GPGraphics.TranslateTransform(-X, -Y);
  GPFontSize := TitleFont.Size;
  GPFont := TGPFont.Create(TitleFont.Name, GPFontSize, FontStyleRegular, UnitPixel);

  GPBrush := TGPSolidBrush.Create(ColorRefToARGB(TitleFont.Color));
  try
    ParentGraph.GPGraphics.DrawString(Title, -1, GPFont, MakePoint(dX , dY), GPBrush);
  finally
    ParentGraph.GPGraphics.ResetTransform;
    GPBrush.Free;
    GPFont.Free;
  end;
end;

procedure TGPGraphAxis.ResetScale;
begin
  FAxisScale := asNormal;
  FMin:=0;
  FMax:=10;
  FRange:=FMax-FMin;
  FLogRange:=1;
  FAutoMin:=true;
  FAutoMax:=true;
  FAutoStep:=true;
  FAutoOrigin:=true;
  AxisChanged(Self);
end;

procedure TGPGraphAxis.Draw(Pos: Integer; Reverse: Boolean);
begin
  if not FVisible then
    Exit;

  DrawAxisLine(Pos);
  if FTicks then
    DrawAxisTicks(Pos, Reverse);
  if FLabels and not Reverse then
    DrawAxisLabels(Pos);
  if not Reverse then
    DrawTitle(Pos);
end;

function TGPGraphAxis.GetOriginMargin: integer;
begin
  if FOrientation = aoHorizontal then
    Result:=OriginPos-FAxisBounds.Left
  else
    Result:=FAxisBounds.Bottom-OriginPos;
end;

function TGPGraphAxis.GetPixelsPerUnit: double;
begin
  Result:=FPlotSize/FRange;
end;

function TGPGraphAxis.GetValuePerPixel:double;
begin
  Result:=FRange/FPlotSize;
end;

{TGPXYGraph}
constructor TGPXYGraph.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  Randomize;
  Width := 300;
  Height := 200;
  Color := clWhite;
  Align:=alNone;
  //ParentColor := true;  2019/12/26 폜
  FBackColor := clWhite;
  FSeriesList:=TList.Create;
  FAllowZoom := True;
  FTitle:=TStringList.Create;
  FTitle.OnChange:=StyleChanged;
//  FTitle.Add('xyGraph');
  FTitleAlignment:=taLeftJustify;
  FTitleFont:=TFont.Create;
  with FTitleFont do begin
    Name:='Arial';
    Color:=clBlack;
    Size:=12;
  end;
  FPiSymbol:='Pi';
  FTitleFont.OnChange:=StyleChanged;

 // FWorkRect:=ClientRect;
  FWorkRect:=Rect(0, 0, 300, 200);
  FPlotArea:=FWorkRect;
  with FPlotArea do
  begin
    FPlotWidth:=Right-Left;
    FPlotHeight:=Bottom-Top;
  end;
  FGridPen:=TPen.Create;
  FGridPen.Style:=psSolid;
  FGridPen.Color:=clGray;
  FGridPen.OnChange:=StyleChanged;
  FGridStyle := gsNone;
  FGridThickness := 1.0;
  FSubGrid := False;
  FSubGridPen:=TPen.Create;
  FSubGridPen.Color:=clGray;
  FSubGridPen.Style:=psSolid;
  FSubGridPen.OnChange:=StyleChanged;
  FSubGridThickness := 1.0;
  FXAxis := TGPGraphAxis.Create(Self);
  FXAxis.Orientation := aoHorizontal;
  FYAxis := TGPGraphAxis.Create(Self);
  FYAxis.FOrientation := aoVertical;
  Doublebuffered := True;
  FAxesFront := True;
end;

procedure TGPXYGraph.Init;
begin
  FAutoAspectRatio := False;
  FFrame := False;
  FTitleAlignment:=taLeftJustify;
  with FTitleFont do begin
    Name:='Arial';
    Color:=clBlack;
    Size:=12;
  end;
  FPiSymbol:='Pi';
  FGridPen.Style:=psSolid;
  FGridPen.Width := 1;
  FGridPen.Color:=clGray;
  FGridStyle := gsNone;
  FGridThickness := 1.0;
  FSubGrid := False;
  FSubGridPen.Color:=clGray;
  FSubGridPen.Width := 1;
  FSubGridPen.Style:=psSolid;
  FSubGridThickness := 1.0;
  FTitle.Text := '';
  FXAxis.Init;
  FYAxis.Init;
end;

destructor TGPXYGraph.Destroy;
var
  i : integer;
begin
  if Assigned(FGPGraphics) then
    FreeAndNil(FGPGraphics);
  FTitle.Free;
  FGridPen.Free;
  FSubGridPen.Free;
  FTitleFont.Free;
  FXAxis.Free;
  FYAxis.Free;
  i :=  SeriesCount -1;
  while i >= 0 do
  begin
    Series[i].Free;
    Dec(i);
  end;
  FSeriesList.Free;
  inherited Destroy;
end;

procedure TGPXYGraph.Error(const Msg: string);
begin
  raise EGPXYGraphError.Create(Msg);
end;

procedure TGPXYGraph.Loaded;
begin
  inherited Loaded;
  with XAxis do
  begin
    SetAxisRange(Min, Max);
    SetTitleAngle(TitleAngle);
  end;
  with YAxis do
  begin
    SetAxisRange(Min, Max);
    SetTitleAngle(TitleAngle);
  end;
//  if Parent <> nil then
//    AdjustGraphArea;
//    ArrangeAxes;
end;

procedure TGPXYGraph.GetChildren(Proc: TGetChildProc; Root: TComponent);
var
  I: Integer;
  ASeries: TGPXYGraphSeries;
begin
  inherited GetChildren(Proc, Root);
  for I := 0 to SeriesCount - 1 do
  begin
    ASeries := Series[I];
//    if ASeries.ParentGraph.Owner = Root then
      Proc(ASeries);
  end;
end;

procedure TGPXYGraph.SetGridStyle(const Value: TGraphGridStyle);
begin
  if Value <> FGridStyle then
  begin
    FGridStyle := Value;
    { gsNone ̏ꍇ SubGrid NA}
    if FGridStyle = gsNone then
      FSubGrid := false;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetGridThickness(const Value: Single);
begin
  if FGridThickness <> Value then
  begin
    FGridThickness := Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetSubGrid(const Value: boolean);
begin
  if Value and (FGridStyle = gsNone) then Exit;
  if Value <> FSubGrid then
  begin
    FSubGrid := Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetGridPen(const Value: TPen);
begin
  FGridPen.Assign(Value);
end;

procedure TGPXYGraph.SetSubGridPen(const Value: TPen);
begin
  FSubGridPen.Assign(Value);
end;

procedure TGPXYGraph.SetSubGridThickness(const Value: Single);
begin
  if FSubGridThickness <> Value then
  begin
    FSubGridThickness := Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetPiSymbol(const Value: string);
begin
  if Value <> FPiSymbol then
  begin
    FPiSymbol:=Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetTitle(const Value: TStringList);
begin
  FTitle.Assign(Value);
end;

procedure TGPXYGraph.SetTitleFont(const Value: TFont);
begin
  FTitleFont.Assign(Value);
end;

procedure TGPXYGraph.SetTitleAlignment(const Value: TAlignment);
begin
  if Value <> FTitleAlignment then
  begin
    FTitleAlignment:=Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetAxesFront(const Value: boolean);
begin
  if Value <> FAxesFront then
  begin
    FAxesFront := Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetBackColor(const Value: TColor);
begin
  if Value <> FBackColor then
  begin
    FBackColor:=Value;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetFrame(const Value: boolean);
begin
   if Value <> FFrame then
   begin
     FFrame := Value;
     StyleChanged(Self);
   end;
end;

procedure TGPXYGraph.SetAutoAspectRatio(const Value: boolean);
begin
  if Value <> FAutoAspectRatio then
  begin
    FAutoAspectRatio := Value;
    if FAutoAspectRatio then
      FAspectRatioFixed := False;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetAspectRatioFixed(const Value: boolean);
begin
  if Value <> FAspectRatioFixed then
  begin
    FAspectRatioFixed := Value;
    if FAspectRatioFixed then
      FAutoAspectRatio := False;
    StyleChanged(Self);
  end;
end;

procedure TGPXYGraph.SetAspectRatio(const Value: double);
begin
  if Value <> FAspectRatio then
  begin
    FAspectRatio := Value;
    FAutoAspectRatio := False;
    FAspectRatioFixed := True;
    StyleChanged(Self);
  end;
end;

function TGPXYGraph.GetSeriesCount: integer;
begin
  Result:=FSeriesList.Count;
end;

function TGPXYGraph.IsFreeColor(Value: TColor): boolean;
var
  i: integer;
begin
  Result:=true;
  for i:=0 to SeriesCount-1 do
    if Series[i].SeriesColor = Value then
    begin
      Result:=false;
      break;
    end;
end;

function TGPXYGraph.GetSeries(Index: integer): TGPXYGraphSeries;
begin
  //if Index < GetSeriesCount then
  Result := TGPXYGraphSeries(FSeriesList[Index]);
end;

procedure TGPXYGraph.SetSeries(Index: integer; ASeries: TGPXYGraphSeries);
begin
  FSeriesList[Index]:=ASeries;
end;

function TGPXYGraph.GetFreeColor: TColor;
var
  k, j, Count, Level: integer;
begin
  for k:=0 to MaxDefaultColors-1 do
  begin
    Result:=ColorPalette[k];
    if IsFreeColor(Result) then
      Exit;
  end;
  Level:=1;
  while true do
  begin
    for k:=0 to MaxDefaultColors-1 do
    begin
      Result:=ColorPalette[k];
      Count:=0;
      for j:=0 to SeriesCount-1 do
        if Series[j].SeriesColor = Result then
          Inc(Count);
      if Count = Level then
        Exit;
    end;
    Inc(Level);
  end;
end;


function TGPXYGraph.MaxXValue: double;
var
  I: integer;
  Temp: double;
  FirstData: boolean;
begin
  Result := XAxis.Max; {defalut}
  FirstData:=true;
  for I := 0 to SeriesCount -1 do
  begin
    if not Series[i].Active then Continue;
    if Series[i].DataCount = 0 then Continue;

    Temp := Series[i].MaxXValue;
    if FirstData then
    begin
      Result := Temp;
      FirstData := false;
      Continue;
    end;
    if Result < Temp then
        Result := Temp;
  end;
end;

function TGPXYGraph.MaxYValue: double;
var
  I: integer;
  Temp: double;
  FirstData: boolean;
begin
  Result := YAxis.Max; {defalut}
  FirstData:=true;
  for I := 0 to SeriesCount -1 do
  begin
    if not Series[i].Active then Continue;
    if Series[i].DataCount = 0 then Continue;
    Temp := Series[i].MaxYValue;
    if FirstData then
    begin
      Result := Temp;
      FirstData := false;
      Continue;
    end;
    if Result < Temp then
        Result := Temp;
  end;
end;

function TGPXYGraph.MinXValue: double;
var
  I: integer;
  Temp: double;
  FirstData: boolean;
begin
  Result := XAxis.Min; {defalut}
  FirstData:=true;
  for I := 0 to SeriesCount -1 do
  begin
    if not Series[i].Active then Continue;
    Temp := Series[i].MinXValue;
    if FirstData then
    begin
      Result := Temp;
      FirstData := false;
      Continue;
    end;
    if Result > Temp then
        Result := Temp;
  end;
end;

function TGPXYGraph.MinYValue: double;
var
  I: integer;
  Temp: double;
  FirstData: boolean;
begin
  Result := YAxis.Min; {defalut}
  FirstData:=true;
  for I := 0 to SeriesCount -1 do
  begin
    if not Series[i].Active then Continue;
    Temp := Series[i].MinYValue;
    if FirstData then
    begin
      Result := Temp;
      FirstData := false;
      Continue;
    end;
    if Result > Temp then
        Result := Temp;
  end;
end;

procedure TGPXYGraph.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  {2019/11/23 inheritedŏɌĂ}
  inherited MouseDown(Button, Shift, X, Y);
{
  if (ssDouble in Shift) then
    Exit;
}
  if AllowZoom and (Button = mbLeft) and not (ssDouble in Shift)  then
    StartZooming(Point(X, Y));
end;

procedure TGPXYGraph.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  {2019/11/23 inheritedŏɌĂ}
  inherited MouseMove(Shift, X, Y);
  if FZooming then
    ZoomingTo(Point(X, Y));
end;

procedure TGPXYGraph.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  {2019/11/23 inheritedŏɌĂ}
  inherited MouseUp(Button, Shift, X, Y);
  if FZooming then
    EndZooming(Point(X, Y));
end;

{ZoomJn}
procedure TGPXYGraph.StartZooming(Pt: TPoint);
var
  R: TRect;
begin
  if PtInRect(ClientRect, Pt) then
  begin
    FZooming:=true;
    FAnchorPt:= Pt;
    FDragPt := Pt;
    R.TopLeft := ClientToScreen(Point(0, 0));
    R.BottomRight := ClientToScreen(Point(Width, Height));
    ClipCursor(@R);
//    Canvas.Pen.Color := BackColor;
  end;
end;

procedure TGPXYGraph.ZoomingTo(Pt: TPoint);
var
  DragRect: TRect;
begin
  {Â`}
  DragRect.TopLeft := FAnchorPt;
  DragRect.BottomRight := FDragPt;
  DrawZoomingRect(DragRect);
  {V``}
  FDragPt := Pt;
  DragRect.BottomRight := FDragPt;
  DrawZoomingRect(DragRect);
end;

procedure TGPXYGraph.EndZooming(Pt: TPoint);
var
  DragRect: TRect;
  DeltaX, DeltaY: Integer;
begin
  FZooming := False;
  ClipCursor(nil);
  if (Pt.X <> FDragPt.X) or (Pt.Y <> FDragPt.Y) then
    ZoomingTo(Pt);
  FZooming := False;
  {Â`}
  DragRect.TopLeft := FAnchorPt;
  DragRect.BottomRight := FDragPt;
  DrawZoomingRect(DragRect);
  {ォEɃo[oh`ꂽƂɂ͊gY[}
  {t̓ɑ΂ẮAY[}
  DeltaX := FDragPt.X - FAnchorPt.X;
  DeltaY := FDragPt.Y - FAnchorPt.Y;
  if (DeltaX > 5) and (DeltaY > 5) then
    ZoomRect(DragRect)
  else if (DeltaX < -5) and (DeltaY < -5) then
    UndoZoom;
end;

procedure TGPXYGraph.DrawZoomingRect(const Rect: TRect);
begin
  //******************************************
  with Canvas do
  begin
    Pen.Width := 1;
    Pen.Mode := pmXor;
    Pen.Color := BackColor;
    Brush.Style := bsClear;
    with Rect do
      Rectangle(Left, Top, Right, Bottom);
  end;
end;

//2017/09/14:  Aj[Vx߂ɏCB
//DoubleBuffered := True ƂĂB
procedure TGPXYGraph.Paint;
begin
  FGPGraphics := TGPGraphics.Create(Canvas.Handle);
  try
    FGPGraphics.SetSmoothingMode(SmoothingModeAntiAlias);
    DrawGraph(FGPGraphics);
  finally
    FGPGraphics.Free;
    FGPGraphics := nil;
  end;
end;

procedure TGPXYGraph.ReadState(Reader: TReader);
begin

  if not FInitialized then
  begin
    FInitialized := True;
    FXAxis.FMin := 0.0;
    FXAxis.FMax := 0.0;
    FYAxis.FMin := 0.0;
    FYAxis.FMax := 0.0;
  end;

  inherited;
end;

procedure TGPXYGraph.StyleChanged(Sender: TObject);
begin
  Invalidate;
end;

procedure TGPXYGraph.UpdateClipLimits(ASeries: TGPXYGraphSeries);
begin
  if ASeries = nil then Exit;
  ASeries.FGraphXmin := XAxis.Min;
  ASeries.FGraphXmax := XAxis.Max;
  ASeries.FGraphYmin := YAxis.Min;
  ASeries.FGraphYmax := YAxis.Max;
end;

procedure TGPXYGraph.SetPlotArea;
begin
  with FPlotArea do
  begin
    Left:=XAxis.MinPos;//-1;
    Right:=XAxis.MaxPos+1;
    Top:=YAxis.MaxPos;//-1;
    Bottom:=YAxis.MinPos+1;
    FPlotWidth:=Right-Left;
    FPlotHeight:=Bottom-Top;
  end;
end;

procedure TGPXYGraph.ArrangeAxes;
var
  NewOriginPos, Margin, AxisSize: integer;
  R: TRect;
begin
  R:=FWorkRect;
  Margin := 0;
  with YAxis do
  begin
    R.Bottom:=R.Bottom-XAxis.GetTitleHeight(tlFixed);
    AxisBounds:=R;
    AxisSize:=XAxis.AxisHeight;
    if GetOriginMargin<AxisSize then {}[WvZ}
    begin
      NewOriginPos:=R.Bottom-AxisSize;
      if NewOriginPos<=MaxPos then
        Margin:=AxisBounds.Bottom -MinPos
      else if not Inverted and (Max <> Origin) then
        Margin:=Round(MinPos-((Max-Min)*NewOriginPos-(Origin-Min)*MaxPos)/(Max-Origin))
      else if Origin <> Min then
        Margin:=Round(MinPos-((Max-Min)*NewOriginPos-(Max-Origin)*MaxPos)/(Origin-Min));
      if Margin >  0 then
      begin
        R.Bottom:=R.Bottom - Margin;
        AxisBounds:=R;  {Vɐݒ肵}
      end;
    end
    else if (AxisHeight-GetOriginMargin)<AxisSize then {}[WvZ}
    begin
      NewOriginPos:=MaxPos;
      if NewOriginPos>=MinPos then
        Margin:=MaxPos - AxisBounds.Top
      else if not Inverted and (Min <> Origin) then
        Margin:=Round(((Max-Min)*NewOriginPos-(Max-Origin)*MinPos)/(Origin-Min)-MaxPos)
      else if Origin <> Max then
        Margin:=Round(((Max-Min)*NewOriginPos-(Origin-Min)*MinPos)/(Max-Origin)-MaxPos);
      if Margin > 0 then
      begin
        R.Top:=R.Top + Margin;
        AxisBounds:=R;  {Vɐݒ肵}
      end;
    end;
  end;{of with}

  with XAxis do
  begin
    R.Left:=R.Left+YAxis.GetTitleWidth(tlFixed);
    AxisBounds:=R;
//    YAxis.AxisBounds:=R;
    AxisSize:=YAxis.AxisWidth;
    if GetOriginMargin<AxisSize then {}[WvZ}
    begin
      NewOriginPos:=R.Left+AxisSize;
      if NewOriginPos>=MaxPos then
        Margin:=MinPos - AxisBounds.Left
      else if not Inverted  and (Max <> Origin) then
        Margin:=Round(((Max-Min)*NewOriginPos-(Origin-Min)*MaxPos)/(Max-Origin)-MinPos)
      else if Origin <> Min then
        Margin:=Round(((Max-Min)*NewOriginPos-(Max-Origin)*MaxPos)/(Origin-Min)-MinPos);
      if Margin > 0 then
      begin
        R.Left:=R.Left + Margin;
        AxisBounds:=R;  {Vɐݒ肵}
      end;
    end
    else if (AxisWidth-GetOriginMargin)<AxisSize then {}[WvZ}
    begin
      NewOriginPos:=MaxPos;
      if NewOriginPos<=MinPos then
        Margin:=AxisBounds.Right - MaxPos
      else if not Inverted and (Min <> Origin) then
        Margin:=Round(MaxPos-((Max-Min)*NewOriginPos-(Max-Origin)*MinPos)/(Origin-Min))
      else if Origin <> Max then
        Margin:=Round(MaxPos-((Max-Min)*NewOriginPos-(Origin-Min)*MinPos)/(Max-Origin));
      if Margin > 0 then
      begin
        R.Right:=R.Right - Margin;
        AxisBounds:=R;
      end;
    end;
  end;

  SetPlotArea;
end;

procedure TGPXYGraph.ArrangeFrame;
var
  R: TRect;
begin
  R:=FWorkRect;
  with YAxis do
  begin
    R.Bottom:=R.Bottom-XAxis.GetTitleHeight(tlFixed)-XAxis.AxisHeight;
    AxisBounds:=R;
  end;{of with}

  with XAxis do
  begin
    R.Left:=R.Left+YAxis.GetTitleWidth(tlFixed)+YAxis.AxisWidth;
    AxisBounds:=R;
  end;

  SetPlotArea;

end;

procedure TGPXYGraph.AdjustGraphArea;
var
  TitleHeight: integer;
  Size: TSize;
  Ratio, CurrAspectRatio: double;
begin
  FWorkRect := ClientRect;
  TitleHeight := 0;
  if Title.Text <> '' then
  begin
    Size := GetTextSize(GPGraphics, FTitle[0], TitleFont);
    TitleHeight := Size.cy * FTitle.Count;
  end;
  FWorkRect.Top := ClientRect.Top + TitleHeight;

  //Frame ̂߂ɒǉ@2017/01/13
  if Frame  then
    ArrangeFrame
  else
    ArrangeAxes;

  with FPlotArea do
  begin
    if Right <> Left then
      CurrAspectRatio := (Bottom -Top)/(Right - Left)
    else
      CurrAspectRatio := 1000;{dummy}
  end;
  if FAutoAspectRatio then
    Ratio := (YAxis.Max - YAxis.Min)/(XAxis.Max - XAxis.Min)
  else if FAspectRatioFixed then
    Ratio := FAspectRatio
  else
    Exit;

  if Ratio < CurrAspectRatio then
    InflateRect(FWorkRect, 0, -(FPlotHeight - Round(FPlotWidth * Ratio)) div 2)
  else
    InflateRect(FWorkRect, -(FPlotWidth-Round(FPlotHeight / Ratio)) div 2, 0);
  //Frame ̂߂ɒǉ@2017/01/13
  if Frame  then
    ArrangeFrame
  else
    ArrangeAxes;
end;

procedure TGPXYGraph.DrawGraph(g: TGPGraphics);
begin
  AdjustGraphArea;
  DrawBackground(g);
  DrawTitle(g);
  if AxesFront then
  begin
    DrawSeries(g);
    DrawGridLines(g);
    if Frame then
      DrawFrame(g)
    else
      DrawAxes(g);
  end
  else
  begin
    DrawGridLines(g);
    if Frame then
      DrawFrame(g)
    else
      DrawAxes(g);
    DrawSeries(g);
  end;
  if Assigned(FOnAfterDraw) then
    OnAfterDraw(Self);

end;

procedure TGPXYGraph.DrawTitle(g: TGPGraphics);
var
  GPFontSize: Single;
  GPFont: TGPFont;
  GPBrush: TGPSolidBrush;
  GPStrFmt: TGPStringFormat;
  GPRect: TGPRectF;
begin
  GPFontSize := TitleFont.Size;
  GPStrFmt := TGPStringFormat.Create;
  GPFont := TGPFont.Create(TitleFont.Name, GPFontSize, FontStyleRegular, UnitPixel);
  GPBrush := TGPSolidBrush.Create(ColorRefToARGB(TitleFont.Color));
  try
    case FTitleAlignment of
      taLeftJustify:
        GPStrFmt.SetAlignment(TStringAlignment(0));
      taCenter:
        GPStrFmt.SetAlignment(TStringAlignment(1));
      taRightJustify:
        GPStrFmt.SetAlignment(TStringAlignment(2));
    end;
    GPRect := MakeRect(0.0, 0.0, Width, Height);
    g.DrawString(Title.Text, -1, GPFont, GPRect, GPStrFmt, GPBrush);
  finally
    GPStrFmt.Free;
    GPFont.Free;
    GPBrush.Free;
  end;
end;

procedure TGPXYGraph.DrawAxes(g: TGPGraphics);
begin
  XAxis.Draw(YAxis.OriginPos, False);
  YAxis.Draw(XAxis.OriginPos, False);
end;

procedure TGPXYGraph.DrawFrame(g: TGPGraphics);
var
  OnAxis: Boolean;
begin
  with XAxis do
  begin
    OnAxis := FLabelsOnAxis;
    FLabelsOnAxis := True;
    Draw(YAxis.FStartPos, False);
    Draw(YAxis.FEndPos, True);
    if not ((YAxis.Origin = YAxis.Min) or (YAxis.Origin = YAxis.Max))  then
      DrawAxisLine(YAxis.OriginPos);
    FLabelsOnAxis := OnAxis;
  end;
  with YAxis do
  begin
    OnAxis := FLabelsOnAxis;
    FLabelsOnAxis := True;
    Draw(XAxis.FStartPos, False);
    Draw(XAxis.FEndPos, True);
    if not (XAxis.Origin = XAxis.Min) or (XAxis.Origin = XAxis.Max) then
      DrawAxisLine(XAxis.OriginPos);
    FLabelsOnAxis := OnAxis;
  end;
end;

procedure TGPXYGraph.DrawBackground(g: TGPGraphics);
var
  GPBrush: TGPSolidBrush;
  GPRect: TGPRectF;
begin
  GPBrush := TGPSolidBrush.Create(ColorRefToARGB(Color));
  try
    GPRect := MakeRect(0.0, 0.0, Width, Height);
    g.FillRectangle(GPBrush, GPRect);
    GPBrush.SetColor(ColorRefToARGB(BackColor));
    with PlotArea do
      GPRect := MakeRect(Left+0.5, Top+0.5, Right - Left+0.5, Bottom - Top+0.5);
    g.FillRectangle(GPBrush, GPRect);
  finally
    GPBrush.Free;
  end;
end;

procedure TGPXYGraph.DrawGridLines(g: TGPGraphics);
var
  i, X ,Y, X1, X2, Y1, Y2: integer;
  Pen, SubPen: TGPPen;
begin
  if FGridStyle = gsNone then
    Exit;

  Pen := CreateGPPen(GridPen, FGridThickness);
  try
    {̃Obh`}
    if FGridStyle in [gsHorizontal, gsBoth] then
    begin
      X1:=XAxis.StartPos;
      X2:=XAxis.EndPos;
      for i:=0 to YAxis.TickDataList.Count-1 do
      begin
        with YAxis do
          Y:=GetValuePos(TickDataList[i]);
        g.DrawLine(Pen, X1, Y, X2, Y);
      end;
      if FSubGrid then
      begin
        SubPen := CreateGPPen(SubGridPen, FSubGridThickness);
        try
          X1:=XAxis.MinPos;
          X2:=XAxis.MaxPos;
          for i:=0 to YAxis.SubTickDataList.Count-1 do
          begin
            with YAxis do
              Y:=GetValuePos(SubTickDataList[i]);
            g.DrawLine(SubPen, X1, Y, X2, Y);
          end;
        finally
          SubPen.Free;
        end;
      end;
    end;
    {c̃Obh`}
    if FGridStyle in [gsVertical, gsBoth] then
    begin
      Y1:=YAxis.StartPos;
      Y2:=YAxis.EndPos;
      for i:=0 to XAxis.TickDataList.Count-1 do
      begin
        with XAxis do
          X:=GetValuePos(TickDataList[i]);
        g.DrawLine(Pen, X, Y1, X, Y2);
      end;
      if FSubGrid then
      begin
        SubPen := CreateGPPen(SubGridPen, FSubGridThickness);
        try
          Y1:=YAxis.MinPos;
          Y2:=YAxis.MaxPos;
          for i:=0 to XAxis.SubTickDataList.Count-1 do
          begin
            with XAxis do
              X:=GetValuePos(SubTickDataList[i]);
            g.DrawLine(SubPen, X, Y1, X, Y2);
          end;
        finally
          SubPen.Free;
        end;
      end;
    end;
  finally
    Pen.Free;
  end;
end;

procedure TGPXYGraph.DrawSeries(g: TGPGraphics);
var
  i: integer;
begin
  if SeriesCount=0 then
    Exit;
  for i:=0 to SeriesCount-1 do
  begin
    UpdateClipLimits(Series[i]);
    with Series[i] do
    begin
      if Active then
        Draw(g);
    end;
  end;
end;

procedure TGPXYGraph.AddSeries(ASeries: TGPXYGraphSeries);
begin
  if ASeries <> nil then
    ASeries.ParentGraph := Self;     //2018/01/15  F
end;

procedure TGPXYGraph.RemoveSeries(ASeries: TGPXYGraphSeries);
begin
  if ASeries <> nil then
    ASeries.ParentGraph := nil;      //2018/01/15  F
end;

procedure TGPXYGraph.RemoveAllSeries;
begin
  while SeriesCount > 0 do
    RemoveSeries(Series[0]);
end;

procedure TGPXYGraph.DeleteAllSeries;
var
  Temp:  TGPXYGraphSeries;
begin
  while SeriesCount > 0 do
  begin
    Temp := Series[0];
    RemoveSeries(Temp);
    Temp.Free;
  end;
end;

function TGPXYGraph.CreateGPPen(APen: TPen; AThickness: Single): TGPPen;
var
  Index: Integer;
begin
  //Pen.Width 1ȊȌꍇ́AÂt@CłB
  //LineTfichness 𒲐B
  if APen.Width > 1 then
    AThickness := APen.Width;
  Result := TGPPen.Create(ColorRefToARGB(APen.Color), AThickness);
  if APen.Style = psClear then
    Result.SetColor(MakeColor(0, 0, 0, 0));  //Ȑ
  if APen.Style < psClear then
     Index := Integer(APen.Style);
  Result.SetDashStyle(DashStyle(Index));
end;

function TGPXYGraph.CreateGraphBitmap: TBitmap;
begin
  Result := TBitmap.Create;
  Result.Width:=ClientWidth;
  Result.Height:=ClientHeight;
  FGPGraphics := TGPGraphics.Create(Result.Canvas.Handle);
  try
    FGPGraphics.SetSmoothingMode(SmoothingModeAntiAlias);
    DrawGraph(FGPGraphics);
  finally
    FGPGraphics.Free;
    FGPGraphics := nil;
  end;
end;

function TGPXYGraph.CreateGraphGifImage: TGifImage;
var
  Bitmap: TBitmap;
begin
  Result := TGifImage.Create;
  Bitmap := CreateGraphBitmap;
  try
    Result.Assign(Bitmap);
  finally
    Bitmap.Free;
  end;
end;

function TGPXYGraph.CreateGraphPNGImage: TPNGImage;
var
  Bitmap: TBitmap;
begin
  Result := TPNGImage.Create;
  Bitmap := CreateGraphBitmap;
  try
    Result.Assign(Bitmap);
  finally
    Bitmap.Free;
  end;

end;

procedure TGPXYGraph.CopyToClipboardBitmap;
var
  Bitmap: TBitmap;
begin
  Bitmap := CreateGraphBitmap;
  try
    ClipBoard.Assign(Bitmap);
  finally
    Bitmap.Free;
  end;
end;

procedure TGPXYGraph.CopyToClipboardPngImage;
var
  png: TPngImage;
begin
  png := CreateGraphPNGImage;
  try
    CopyPngImageToClipboard(png);
  finally
    png.Free;
  end;
end;

//ȉ̃R[hPNG - Image mit Transparenz in Clipboard 
(*
procedure TGPXYGraph.CopyToClipboardPngImage;
var
  png: TPngImage;
  DataStream: TMemoryStream;
  Data: Pointer;
  DataHandle: THandle;
  ClipboardFormat: Integer;
begin
  png := CreateGraphPNGImage;
  try
    ClipboardFormat := RegisterClipboardFormat('PNG');
    DataStream := TMemoryStream.Create;
    try
      png.SaveToStream(DataStream);
      DataHandle := GlobalAlloc(GMEM_DDESHARE, DataStream.Size);
      Data := GlobalLock(DataHandle);
      try
        DataStream.Position := 0;
        DataStream.ReadBuffer(Data^, DataStream.Size);
      finally
        GlobalUnlock(DataHandle);
      end;
      Clipboard.SetAsHandle(ClipboardFormat, DataHandle);
    finally
      DataStream.Free;
    end;
  finally
    png.Free;
  end;
end;
*)

procedure TGPXYGraph.SaveToMetafile(const Filename: string);
var
  Metafile: TMetafile;
begin
  Metafile := CreateGraphMetafile(false);
  try
    Metafile.SaveToFile(Filename)
  finally
    Metafile.Free;
  end;
end;

procedure TGPXYGraph.SaveToMetafileEnh(const Filename: string);
var
  Metafile: TMetafile;
begin
  Metafile := CreateGraphMetafile(true);
  try
    Metafile.SaveToFile(Filename)
  finally
    Metafile.Free;
  end;
end;

procedure TGPXYGraph.SaveGraphToBitmapFile(const FileName : string);
var
  Bitmap: TBitmap;
begin
  Bitmap := CreateGraphBitmap;
  try
    Bitmap.SaveToFile(FileName);
  finally
    Bitmap.Free;
  end;
end;

procedure TGPXYGraph.SaveGraphToGifFile(const Filename: string);
var
  gif: TGifImage;
begin
  gif := CreateGraphGifImage;
  try
    gif.SaveToFile(FileName);
  finally
    gif.Free;
  end;
end;

procedure TGPXYGraph.SaveGraphToPngFile(const Filename: string);
var
  png: TPngImage;
begin
  png := CreateGraphPNGImage;
  try
    png.SaveToFile(FileName);
  finally
    png.Free;
  end;
end;

function TGPXYGraph.CreateGraphMetafile(EMF: boolean): TMetafile;
var
  MetafileCanvas: TMetafileCanvas;
begin
  Result := TMetafile.Create;
  Result.Enhanced:=EMF;
  Result.Width:=ClientWidth;
  Result.Height:=ClientHeight;
  MetafileCanvas := TMetafileCanvas.Create(Result, 0);
  FMetafiling:=true;
  try
    FGPGraphics := TGPGraphics.Create(MetafileCanvas.Handle);
    try
      FGPGraphics.SetSmoothingMode(SmoothingModeAntiAlias);
      DrawGraph(FGPGraphics);
    finally
      FGPGraphics.Free;
      FGPGraphics := nil;
    end;
  finally
    FMetafiling := false;
    MetafileCanvas.Free;
    // FTargetCanvas ɖ߂
  end;                                  ;
end;

procedure TGPXYGraph.CopyToClipboardMetafile(EMF: boolean);
var
  Metafile: TMetafile;
begin
  Metafile := CreateGraphMetafile(EMF);
  try
    ClipBoard.Assign(Metafile);
  finally
    Metafile.Free;
  end;
end;

{Zoom ֘A葱 2001/03/31}
procedure TGPXYGraph.UndoZoom;
begin
  if FZoomed then
  begin
    try
      with FZoomingData do
      begin
        XAxis.SetMinMax(XMin, XMax);
        YAxis.SetMinMax(YMin, YMax);
        XAxis.FAutoMax := XAutoMax;
        XAxis.FAutoMin := XAutoMin;
        YAxis.FAutoMax := YAutoMax;
        YAxis.FAutoMin := YAutoMin;
      end;
    finally
      FZoomed := False;
    end;
    {OnUndoZoom CxgnhĂ}
    if Assigned(FOnUndoZoom) then FOnUndoZoom(Self);
    Invalidate;
  end;
end;

procedure TGPXYGraph.ZoomRect(R: TRect);
var
  InterRect, GraphRect: TRect;
  V1, V2: double;
begin
  GraphRect := Rect(XAxis.MinPos, YAxis.MaxPos, XAxis.MaxPos, YAxis.MinPos);
  IntersectRect(InterRect, R, GraphRect);
  if IsRectEmpty(InterRect) then Exit;

  if not FZoomed then
  begin
    with FZoomingData do
    begin
      XMax := XAxis.Max;
      XMin := XAxis.Min;
      YMax := YAxis.Max;
      YMin := YAxis.Min;
      XAutoMax := XAxis.AutoMax;
      XAutoMin := XAxis.AutoMin;
      YAutoMax := YAxis.AutoMax;
      YAutoMin := YAxis.AutoMin;
    end;
    FZoomed := True;
  end;
  V1 := XAxis.GetPosValue(InterRect.Left);
  V2 := XAxis.GetPosValue(InterRect.Right);
  XAxis.SetMinMax(Min(V1, V2), Max(V1, V2));
  V1 := YAxis.GetPosValue(InterRect.Bottom);
  V2 := YAxis.GetPosValue(InterRect.Top);
  YAxis.SetMinMax(Min(V1, V2), Max(V1, V2));
  {OnZoom CxgnhĂ}
  if Assigned(FOnZoom) then FOnZoom(Self);

  Invalidate;
end;

function TGPXYGraph.GetXYPoint(X, Y: Double): TPoint;
begin
  Result := Point(XAxis.GetValuePos(X),YAxis.GetValuePos(Y));
end;

procedure TGPXYGraph.SeriesUp(ASeries: TGPXYGraphSeries);
var
  Index: integer;
begin
  Index := FSeriesList.IndexOf(ASeries);
  if Index > 0 then
  begin
    FSeriesList.Exchange(Index, Index - 1);
    Invalidate;
  end;
end;

procedure TGPXYGraph.SeriesDown(ASeries: TGPXYGraphSeries);
var
  Index: integer;
begin
  Index := FSeriesList.IndexOf(ASeries);
  if Index < SeriesCount - 1 then
  begin
    FSeriesList.Exchange(Index, Index + 1);
    Invalidate;
  end;
end;

function TGPXYGraph.SeriesIndexOf(ASeries: TGPXYGraphSeries): integer;
begin
  Result := FSeriesList.IndexOf(ASeries);
end;

{TGPXYGraphSeries}
constructor TGPXYGraphSeries.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FXValues:=TGraphValueList.Create;
  FYValues:=TGraphValueList.Create;
  FSeparators:=TGraphBooleanList.Create;
  FSeriesColor := clRed;
  FActive := true;
end;

destructor TGPXYGraphSeries.Destroy;
begin
  SetParentGraph(nil);
  FXValues.Free;
  FYValues.Free;
  FSeparators.Free;
  inherited Destroy;
end;

procedure TGPXYGraphSeries.Assign(Source: TPersistent);
begin
  if Source is TGPXYGraphSeries then
  begin
    SeriesColor := TGPXYGraphSeries(Source).SeriesColor;
    Active := TGPXYGraphSeries(Source).Active;
  end
  else
    inherited Assign(Source);
end;

procedure TGPXYGraphSeries.MoveTo(Dest: TGPXYGraph);
begin
  SetParentGraph(Dest);
end;

procedure TGPXYGraphSeries.Error(const Msg: string);
begin
  raise EXYGraphSeriesError.Create(Msg);
end;

procedure TGPXYGraphSeries.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FParentGraph) and (Operation = opRemove) then
    FParentGraph:=nil;
end;

function TGPXYGraphSeries.HasParent: Boolean;
begin
  Result := FParentGraph <> nil;
end;

function TGPXYGraphSeries.GetParentComponent: TComponent;
begin
  Result := FParentGraph;
end;

procedure TGPXYGraphSeries.SetParentComponent(Value: TComponent);
begin
  if Value is TGPXYGraph then
    SetParentGraph(TGPXYGraph(Value))
  else inherited;
end;

procedure TGPXYGraphSeries.SetParentGraph(Value: TGPXYGraph);
begin
  if FParentGraph <> Value then
  begin
    if FParentGraph <> nil then
    begin
      FParentGraph.FSeriesList.Remove(Self);
      FParentGraph.Invalidate;
    end;
    FParentGraph:=Value;
    if Value <> nil then
    begin
      Value.FSeriesList.Add(Self);
      Value.FreeNotification(Self);
      if csDesigning in ComponentState then
      begin
        //Value.UpdateClipLimits(Self);
        SeriesColor := Value.GetFreeColor;
        FillSamples(10);
      end;
(* 2020/02/23 폜@2sڂɈړ
     if SeriesColor = -1 then
        SeriesColor := Value.GetFreeColor;
*)
      Value.Invalidate;
    end;

  end;
end;

procedure TGPXYGraphSeries.Clear;
begin
  XValues.Clear;
  YValues.Clear;
  Separators.Clear;
end;

procedure TGPXYGraphSeries.Delete(Index: integer);
begin
  XValues.Delete(Index);
  YValues.Delete(Index);
  Separators.Delete(Index);
end;

procedure TGPXYGraphSeries.Changed;
begin
  if (FUpdateCount = 0) and (ParentGraph <> nil) then
   ParentGraph.Invalidate;
end;

procedure TGPXYGraphSeries.DeviceChanged(Sender: TObject);
begin
  Changed;
end;

procedure TGPXYGraphSeries.Draw(g: TGPGraphics);
begin
 //Do Nothing
end;

procedure TGPXYGraphSeries.SetActive(const Value: boolean);
begin
  if FActive <> Value then
  begin
    FActive := Value;
    Changed;
  end;
end;

procedure TGPXYGraphSeries.SetSeriesColor(Value: TColor);
begin
  if FSeriesColor <> Value then
  begin
    FSeriesColor := Value;
    Changed;
  end;
end;

function TGPXYGraphSeries.GetXValue(Index: integer): double;
begin
  Result:=XValues[Index];
end;

procedure TGPXYGraphSeries.SetXValue(Index: integer; Value: double);
begin
  XValues[Index]:=Value;
  Changed;
end;

function TGPXYGraphSeries.GetYValue(Index: integer): double;
begin
  Result:=YValues[Index];
end;

procedure TGPXYGraphSeries.SetYValue(Index: integer; Value: double);
begin
  YValues[Index]:=Value;
  Changed;
end;

procedure TGPXYGraphSeries.UpdateValues;
begin
  if (csLoading in ComponentState) then Exit;
  Changed;
  // Nothing to do
end;

function TGPXYGraphSeries.GetSeparatorValue(Index: integer): Boolean;
begin
  Result:=Separators[Index];
end;

procedure TGPXYGraphSeries.SetSeparatorValue(Index: integer; Value: Boolean);
begin
  Separators[Index]:=Value;
  Changed;
end;

function TGPXYGraphSeries.DataCount: integer;
begin
  Result := XValues.Count;
end;

function TGPXYGraphSeries.AddXY(X,Y: double): Integer;
begin
  Result := DataCount - 1;
  XValues.Add(X);
  YValues.Add(Y);
  Separators.Add(False);
  Inc(Result);
  if Assigned(OnAfterAdd) then
    OnAfterAdd(Self, Result);
  //Changed;
end;

function TGPXYGraphSeries.AddSeparator: Integer;
begin
  Result := DataCount - 1;
  if (DataCount > 0) and Separators[Result] then
    Exit;
  XValues.Add(0);
  YValues.Add(0);
  Separators.Add(True);
  Inc(Result);
  if Assigned(OnAfterAdd) then
    OnAfterAdd(Self, Result);
  //Changed;
end;

function TGPXYGraphSeries.MinXValue: double;
begin
  Result := XValues.MinValue;
end;

function TGPXYGraphSeries.MaxXValue: double;
begin
  Result := XValues.MaxValue;
end;

function TGPXYGraphSeries.MinYValue: double;
begin
  Result := YValues.MinValue;
end;

function TGPXYGraphSeries.MaxYValue: double;
begin
  Result := YValues.MaxValue;
end;

procedure TGPXYGraphSeries.FillSamples(Value: integer);
var
  Index, I : integer;
  A, X, Y, dX, Ymean: double;
begin
  if Value > 0 then
  begin
    if ParentGraph = nil then
      Error(Name + ' ͐eOtĂ܂');
    ParentGraph.UpdateClipLimits(Self);
    Index := GetIndex;
    Clear;
    Randomize;
    dX := (FGraphXmax - FGraphXmin)/Value;
    A := (FGraphYmax - FGraphYmin)/2;
    Ymean := (FGraphYmin + FGraphYmax)/2;
    for I:=0 to Value  - 1 do
    begin
      X := FGraphXmin + I*dx;
      Y := Ymean + A * Cos(I*Pi/Value - Index * Pi/3)/2 + A * Random/5;
      AddXY(X, Y);
    end;
  end;
end;

function TGPXYGraphSeries.GetIndex: integer;
begin
  Result := -1;
  if ParentGraph <> nil then
    Result := ParentGraph.FSeriesList.IndexOf(Pointer(Self));
end;

function TGPXYGraphSeries.IsSeparator(const I: Integer): boolean;
begin
  Result := (Separators.Count > 0) and Separators.List[I];
end;

procedure TGPXYGraphSeries.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TGPXYGraphSeries.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then
    UpdateValues;
end;

initialization
  RegisterClasses([TGPXYGraph, TGPXYGraphSeries]);

end.
